# this module defines the tools for modeling

# __all__ = ['trends', 'seasonality', 'dynamic', 'autoReg', 'longSeason', 'builder']

# import pydlm.modeler.trends as trends
# import pydlm.modeler.seasonality as seasonality
# import pydlm.modeler.dynamic as dynamic
# import pydlm.modeler.builder as builder
# import pydlm.modeler.autoReg as autoReg
# import pydlm.modeler.longSeason as longSeason



import unittest
from pydlm.modeler.autoReg import autoReg


class testAutoReg(unittest.TestCase):


    def setUp(self):
        self.data = [0, 1, 2, 3, 0, 1, 2, 3, 0, 1, 2, 3]
        self.ar4 = autoReg(degree=4, name='ar4', padding=0, w=1.0)


    def testFeatureMatrix(self):
        trueFeatures = [[0, 0, 0, 0], [0, 0, 0, 0], [0, 0, 0, 1], [0, 0, 1, 2],
                        [0, 1, 2, 3], [1, 2, 3, 0], [2, 3, 0, 1], [3, 0, 1, 2],
                        [0, 1, 2, 3], [1, 2, 3, 0], [2, 3, 0, 1], [3, 0, 1, 2]]
        actualFeatures = []
        for i in range(12):
            self.ar4.updateEvaluation(i, self.data)
            actualFeatures.append(self.ar4.evaluation.A1.tolist())

        self.assertEqual(actualFeatures, trueFeatures)


if __name__ == '__main__':
    unittest.main()



"""
=================================================================

Code for the base model structure

=================================================================

This piece of code provides the basic model structure for dynamic linear model.
It stores all the necessary components for kalmanFilter and save the results

"""
# dependencies
import numpy as np
import pydlm.base.tools as tl


# define the basic structure for a dlm model
class baseModel:
    """ The baseModel class that provides the basic model structure for dlm. 

    Attributes:
        transition: the transition matrix G
        evaluation: the evaluation F
        noiseVar: the variance of the observation noise
        sysVar: the covariance of the underlying states
        innovation: the incremnent of the latent covariance W
        state: the latent states
        df: the degree of freedom (= number of data points)
        obs: the expectation of the observation
        obsVar: the variance of the observation

    Methods:
        initializeObservation: initialize the obs and obsVar
        validation: validate the matrix dimensions are consistent.
    """


    # define the components of a baseModel
    def __init__(self, transition = None, evaluation = None, noiseVar = None, \
                 sysVar = None, innovation = None, state = None, df = None):
        self.transition = transition
        self.evaluation = evaluation
        self.noiseVar = noiseVar
        self.sysVar = sysVar
        self.innovation = innovation
        self.state = state
        self.df = df
        self.obs = None
        self.obsVar = None

        # a hidden data field used only for model prediction
        self.prediction = __model__()


    # initialize the observation mean and variance
    def initializeObservation(self):
        """ Initialize the value of obs and obsVar

        """
        self.validation()
        self.obs = np.dot(self.evaluation, self.state)
        self.obsVar = np.dot(np.dot(self.evaluation, self.sysVar), self.evaluation.T) \
                      + self.noiseVar


    # checking if the dimension matches with each other
    def validation(self):
        """ Validate the model components are consistent

        """
        # check symmetric
        tl.checker.checkSymmetry(self.transition)
        tl.checker.checkSymmetry(self.sysVar)
        if self.innovation is not None:
            tl.checker.checkSymmetry(self.innovation)

        # check wether dimension match
        tl.checker.checkMatrixDimension(self.transition, self.sysVar)
        if self.innovation is not None:
            tl.checker.checkMatrixDimension(self.transition, self.innovation)
        tl.checker.checkVectorDimension(self.evaluation, self.transition)
        tl.checker.checkVectorDimension(self.state, self.transition)

        
# define an inner class to store intermediate results
class __model__:

    # to store result for prediction
    def __init__(self, step = 0, state = None, obs = None, sysVar = None, obsVar = None):
        self.step = 0
        self.state = state
        self.obs = obs
        self.sysVar = sysVar
        self.obsVar = obsVar



import numpy as np
import unittest
from pydlm.modeler.builder import builder
from pydlm.modeler.trends import trend
from pydlm.modeler.seasonality import seasonality
from pydlm.modeler.dynamic import dynamic
from pydlm.modeler.autoReg import autoReg
from pydlm.modeler.matrixTools import matrixTools as mt


class testBuilder(unittest.TestCase):


    def setUp(self):
        self.data = np.random.rand(10).tolist()
        self.features = np.random.rand(10, 2).tolist()
        self.trend = trend(degree=2, w=1.0)
        self.seasonality = seasonality(period=7, w=1.0)
        self.dynamic = dynamic(self.features, w=1.0)
        self.autoReg = autoReg(degree=3,
                               w=1.0)
        self.builder1 = builder()
        self.builder2 = builder()


    def testInitialization(self):
        self.assertEqual(len(self.builder1.staticComponents), 0)
        self.assertEqual(len(self.builder1.dynamicComponents), 0)
        self.assertEqual(len(self.builder1.automaticComponents), 0)


    def testAddAndDelete(self):
        self.builder1 = self.builder1 + self.trend
        self.assertEqual(len(self.builder1.staticComponents), 1)
        self.assertEqual(len(self.builder1.dynamicComponents), 0)
        self.assertEqual(len(self.builder1.automaticComponents), 0)

        self.builder1 = self.builder1 + self.dynamic
        self.assertEqual(len(self.builder1.staticComponents), 1)
        self.assertEqual(len(self.builder1.dynamicComponents), 1)
        self.assertEqual(len(self.builder1.automaticComponents), 0)

        self.builder1 = self.builder1 + self.seasonality
        self.assertEqual(len(self.builder1.staticComponents), 2)
        self.assertEqual(len(self.builder1.dynamicComponents), 1)
        self.assertEqual(len(self.builder1.automaticComponents), 0)

        self.builder1.delete('seasonality')
        self.assertEqual(len(self.builder1.staticComponents), 1)
        self.assertEqual(len(self.builder1.dynamicComponents), 1)
        self.assertEqual(len(self.builder1.automaticComponents), 0)

        self.assertEqual(self.builder1.staticComponents['trend'], self.trend)

        self.builder1 = self.builder1 + self.autoReg
        self.assertEqual(len(self.builder1.automaticComponents), 1)


    def testInitialize(self):
        self.builder1 = self.builder1 + self.trend + self.dynamic \
                        + self.autoReg

        self.builder1.initialize(data=self.data)
        self.assertAlmostEqual(np.sum(
            np.abs(self.builder1.model.evaluation
                   - mt.matrixAddByCol(mt.matrixAddByCol(
                       self.trend.evaluation,
                       self.dynamic.evaluation),
                        self.autoReg.evaluation))), 0.0)


    def testInitializeEvaluatoin(self):
        self.builder1 = self.builder1 + self.trend + self.dynamic
        self.builder1.dynamicComponents['dynamic'].updateEvaluation(8)
        self.builder1.initialize(data=self.data)
        self.assertAlmostEqual(np.sum(
            np.abs(self.builder1.model.evaluation -
                   mt.matrixAddByCol(self.trend.evaluation,
                                     self.dynamic.evaluation))), 0.0)


    def testUpdate(self):
        self.builder1 = self.builder1 + self.trend + self.dynamic \
                        + self.autoReg

        self.builder1.initialize(data=self.data)
        self.builder1.updateEvaluation(2, self.data)
        self.assertAlmostEqual(np.sum(
            np.abs(self.builder1.model.evaluation
                   - mt.matrixAddByCol(mt.matrixAddByCol(
                       self.trend.evaluation,
                       np.matrix([self.features[2]])),
                                       np.matrix(self.autoReg.evaluation)))), 0.0)


    def testInitializFromBuilder(self):
        self.builder1 = self.builder1 + self.trend + self.dynamic
        self.builder1.dynamicComponents['dynamic'].updateEvaluation(8)
        self.builder1.initialize(data=self.data)

        self.data1 = self.data[8:-1]
        self.builder2.initializeFromBuilder(exported_builder=self.builder1, data=self.data1)

        # make sure builder1 and builder2 are identical.
        self.assertDictEqual(self.builder1.staticComponents,
                             self.builder2.staticComponents)
        self.assertDictEqual(self.builder1.dynamicComponents,
                             self.builder2.dynamicComponents)
        self.assertDictEqual(self.builder1.automaticComponents,
                             self.builder2.automaticComponents)

if __name__ == '__main__':
    unittest.main()



"""
===============================================================

The discount factor tuner for dlm

===============================================================

The modelTuner class provides the tuning functionality for the dlm class.
It makes use of the gradient descent to optimize the discount factor for
each component (jointly) based on the one-day ahead prediction error.

>>> import modelTuner
>>> myTuner = modelTuner()
>>> tunedDLM = myTuner(untunedDLM, maxit=100)

The tunedDLM will be saved in tunedDLM while the untunedDLM remains unchangd.
An alternative way to call this class is via the tuner method within dlm class.

>>> mydlm.tune(maxit=100)

This will permenantly change the discouting factor in mydlm. So if the user
prefer to build a new dlm with the new discount factor without changing the 
original one, one should opt to use the modelTuner class.

"""
from copy import deepcopy
from numpy import array
import logging

class modelTuner:
    """ The main class for modelTuner

    Attributes:
        method: the optimization method. Currently only 'gradient_descent'
                is supported.
        loss:   the optimization loss function. Currently only 'mse' (one-day
                ahead prediction) is supported.
    
    """

    def __init__(self, method='gradient_descent', loss='mse'):

        self.method = method
        self.loss = loss
        self.current_mse = None
        self.err = 1e-4
        self.discounts = None

    def tune(self, untunedDLM, maxit=100, step = 1.0):
        """ Main function for tuning the DLM model.

        Args:
            untunedDLM: The DLM object that needs tuning
            maxit: The maximum number of iteractions for gradient descent.
            step: the moving length at each iteraction.

        Returns:
            A tuned DLM object in unintialized status.
        """
        # make a deep copy of the original dlm
        tunedDLM = deepcopy(untunedDLM)

        if not tunedDLM.initialized:
            tunedDLM.fitForwardFilter()
        discounts = array(tunedDLM._getDiscounts())
        self.current_mse = tunedDLM._getMSE()
        
        # using gradient descent
        if self.method == 'gradient_descent':

            # Disable all info and warning for faster processing.
            log_level = tunedDLM.getLoggingLevel()
            tunedDLM.setLoggingLevel('CRITICAL')

            for i in range(maxit):
                gradient = self.find_gradient(discounts, tunedDLM)
                discounts -= gradient * step
                discounts = list(map(lambda x: self.cutoff(x), discounts))
                tunedDLM._setDiscounts(discounts)
                tunedDLM.fitForwardFilter()
                self.current_mse = tunedDLM._getMSE()

            # Recover logger level
            tunedDLM.setLoggingLevel(log_level)

            if i < maxit - 1:
                tunedDLM._logger.info('Converge successfully!')
            else:
                tunedDLM._logger.warning('The algorithm stops without converging.')
                if min(discounts) <= 0.7 + self.err or max(discounts) >= 1 - 2 * self.err:
                    tunedDLM._logger.info('Possible reason: some discount is too close to 1 or 0.7'
                                          ' (0.7 is smallest discount that is permissible.')
                else:
                    tunedDLM._logger.info('It might require more step to converge.'
                                          ' Use tune(..., maixt = <a larger number>) instead.')

        self.discounts = discounts
        tunedDLM._setDiscounts(discounts, change_component=True)
        return tunedDLM

    def getDiscounts(self):
        """ Get the tuned discounting factors. One for each component (even the
            component being multi-dimensional, only one discounting factor will
            be assigned to one component). Initialized to None.

        """
        return self.discounts

    def find_gradient(self, discounts, DLM):
        if self.current_mse is None:
            self.current_mse = DLM._getMSE()

        gradient = array([0.0] * len(discounts))

        for i in range(len(discounts)):
            discounts_err = discounts
            discounts_err[i] = self.cutoff(discounts_err[i] + self.err)

            DLM._setDiscounts(discounts_err)
            DLM.fitForwardFilter()
            gradient[i] = (DLM._getMSE() - self.current_mse) / self.err

        return gradient

    def cutoff(self, a):
        if a < 0.7:
            return 0.7

        if a >= 1:
            return 0.99999

        return a



import unittest
import numpy as np

from copy import deepcopy
from pydlm.tuner.dlmTuner import modelTuner
from pydlm.modeler.trends import trend
from pydlm.dlm import dlm


class testModelTuner(unittest.TestCase):


    def setUp(self):
        self.mydlm = dlm(np.random.random(100)) + trend(2, discount=0.95)
        self.mytuner = modelTuner()


    def testFind_gradient(self):
        mydlm2 = deepcopy(self.mydlm)
        self.mydlm.fitForwardFilter()
        mydlm2.fitForwardFilter()
        mse0 = mydlm2._getMSE()
        mydlm2._setDiscounts(list(map(lambda x: x + self.mytuner.err,
                                      self.mydlm._getDiscounts())))
        mydlm2.fitForwardFilter()
        mse1 = mydlm2._getMSE()
        expect_gradient = (mse1 - mse0) / self.mytuner.err
        self.assertAlmostEqual(
            expect_gradient, self.mytuner.find_gradient(
                self.mydlm._getDiscounts(), self.mydlm))


if __name__ == '__main__':
    unittest.main()



"""
==============================================================================

Code for builder of a dynamic linear model

==============================================================================

This piece of code provides the basic functionality for constructing the model
of a dlm. It allows flexible modeling by users. User can add to, delete and
view componets of a given dlm. Builder will finally assemble all the components
 to a final big model.

"""
# this class provide all the model building operations for constructing
# customized model

from pydlm.base.baseModel import baseModel
from pydlm.modeler.matrixTools import matrixTools as mt

from copy import deepcopy
import logging
import numpy as np

# The builder will be the main class for construting dlm
# it featues two types of evaluation matrix and evaluation matrix
# The static evaluation remains the same over time which is used to
# record the trend and seasonality.
#
# The dynamic evaluation vector changes over time, it is basically
# the other variables that might have impact on the time series
# We need to update this vector as time going forward


class builder:
    """ The main modeling part of a dynamic linear model. It allows the users to
    custemize their own model. User can add, delete any components like trend
    or seasonality to the builder, or view the existing components. Builder
    will finally assemble all components to a big model for further training
    and inference.

    Attributes:
        model: the model structure from @baseModel, stores all the necessary
        quantities initialized: indicates whether the model has been built
        staticComponents: stores all the static components (trend and
                          seasonality)
        dynamicComponents: stores all the dynamic components
        componentIndex: the location of each component in the latent states
        statePrior: the prior mean of the latent state
        sysVarPrior: the prior of the covariance of the latent states
        noiseVar: the prior of the observation noise
        initialDegreeFreedom: the initial degree of freedom.
        discount: the discounting factor, please refer to @kalmanFilter for
                  more details
        renewTerm: used for aiding the stability of the model. The renewTerm
                   is computed according to the discount fact. When the filter
                   goes over certain steps, the information contribution of the
                   previous data has decayed to minimum. We then ignore those
                   days and refit the time series starting from current -
                   renewTerm. Thus, the effective sample size of the dlm is
                   twice renewTerm. When discount = 1, there will be no
                   renewTerm, since all the information will be passed along.
        renewDiscount: the minimum discount of seasonality component,
                       or if there is no seasonality, this will be the minimum
                       discount of all components. Used for computing
                       renewTerm.

    Methods:
        add: add new component
        ls:  list out all components
        delete: delete a specific component by its name
        initialize: assemble all the component to construt a big model
        updateEvaluation: update the valuation matrix of the big model
    """

    # create members
    def __init__(self, logger=None):

        # the basic model structure for running kalman filter
        self.model = None
        self.initialized = False

        # to store all components. Separate the two as the evaluation
        # for dynamic component needs update each iteration
        # (new) added the third category, which is dynamic but the
        # features can be automatically computed from the data

        # store all the static components, i.e., the evaluation vector
        # do not change over time
        self.staticComponents = {}

        # store all the dynamic components, i.e., the evaluation vector
        # changes over time
        self.dynamicComponents = {}

        # store all the components that are dynamic, but the evaluation
        # vector can be inferred directly from the main data instead
        # of other sources
        self.automaticComponents = {}

        # store the index (location in the latent states) of all components
        # can be used to extract information for each componnet
        self.componentIndex = {}

        # record the prior guess on the latent state and system covariance
        self.statePrior = None
        self.sysVarPrior = None
        self.noiseVar = None
        self.initialDegreeFreedom = 1

        # record the discount factor for the model
        self.discount = None

        # renew term used to indicate the effective length of data, i.e.,
        # for days before this length will have little impact on
        # the current result
        self.renewTerm = -1.0
        self.renewDiscount = None  # used for adjusting renewTerm

        # Allow customized logger to control the level of information printed
        # during job execution. The default log level is 'INFO'.
        if logger is None:
            self._logger = logging.getLogger()
        else:
            self._logger = logger

    # The function that allows the user to add components
    def add(self, component):
        """ Add a new model component to the builder.

        Args:
            component: a model component, any class implements @component class

        """

        self.__add__(component)

    def __add__(self, component):
        if component.componentType == 'dynamic':
            if component.name in self.dynamicComponents:
                raise NameError('Please rename the component to a'
                                + ' different name.')
            self.dynamicComponents[component.name] = component

        if component.componentType == 'autoReg' \
           or component.componentType == 'longSeason':
            if component.name in self.automaticComponents:
                raise NameError('Please rename the component to a'
                                + ' different name.')
            self.automaticComponents[component.name] = component

        if component.componentType == 'trend' \
           or component.componentType == 'seasonality':
            if component.name in self.staticComponents:
                raise NameError('Please rename the component' +
                                ' to a different name.')
            self.staticComponents[component.name] = component

            # we use seasonality's discount to adjust the renewTerm
            if component.componentType == 'seasonality':
                if self.renewDiscount is None:
                    self.renewDiscount = 1.0
                self.renewDiscount = min(self.renewDiscount,
                                         min(component.discount))
        self.initialized = False
        return self

    # print all components to the client
    def ls(self):
        """ List out all the existing components to the model

        """

        if len(self.staticComponents) > 0:
            print('The static components are')
            for name in self.staticComponents:
                comp = self.staticComponents[name]
                print(comp.name + ' (degree = ' + str(comp.d) + ')')
            print(' ')
        else:
            print('There is no static component.')
            print(' ')

        if len(self.dynamicComponents) > 0:
            print('The dynamic components are')
            for name in self.dynamicComponents:
                comp = self.dynamicComponents[name]
                print(comp.name + ' (dimension = ' + str(comp.d) + ')')
            print(' ')
        else:
            print('There is no dynamic component.')
            print(' ')

        if len(self.automaticComponents) > 0:
            print('The automatic components are')
            for name in self.automaticComponents:
                comp = self.automaticComponents[name]
                print(comp.name + ' (dimension = ' + str(comp.d) + ')')
        else:
            print('There is no automatic component.')

    # delete the componet that pointed out by the client
    def delete(self, name):
        """ Delete a specific component from dlm by its name.

        Args:
            name: the name of the component. Can be read from ls()

        """

        if name in self.staticComponents:
            del self.staticComponents[name]
        elif name in self.dynamicComponents:
            del self.dynamicComponents[name]
        elif name in self.automaticComponents:
            del self.automaticComponents[name]
        else:
            raise NameError('Such component does not exisit!')

        self.initialized = False

    # initialize model for all the quantities
    # noise is the prior guess of the variance of the observed data
    # data is used by auto regressor.
    def initialize(self, data=[], noise=1):
        """ Initialize the model. It construct the baseModel by assembling all
        quantities from the components.

        Args:
            noise: the initial guess of the variance of the observation noise.
        """
        if len(self.staticComponents) == 0 and \
           len(self.dynamicComponents) == 0 and \
           len(self.automaticComponents) == 0:

            raise NameError('The model must contain at least' +
                            ' one component')

        # construct transition, evaluation, prior state, prior covariance
        self._logger.info('Initializing models...')
        transition = None
        evaluation = None
        state = None
        sysVar = None
        self.discount = np.array([])

        # first construct for the static components
        # the evaluation will be treated separately for static or dynamic
        # as the latter one will change over time
        currentIndex = 0  # used for compute the index
        for i in self.staticComponents:
            comp = self.staticComponents[i]
            transition = mt.matrixAddInDiag(transition, comp.transition)
            evaluation = mt.matrixAddByCol(evaluation,
                                           comp.evaluation)
            state = mt.matrixAddByRow(state, comp.meanPrior)
            sysVar = mt.matrixAddInDiag(sysVar, comp.covPrior)
            self.discount = np.concatenate((self.discount, comp.discount))
            self.componentIndex[i] = (currentIndex, currentIndex + comp.d - 1)
            currentIndex += comp.d

        # if the model contains the dynamic part, we add the dynamic components
        if len(self.dynamicComponents) > 0:
            self.dynamicEvaluation = None
            for i in self.dynamicComponents:
                comp = self.dynamicComponents[i]
                comp.updateEvaluation(0)
                transition = mt.matrixAddInDiag(transition, comp.transition)
                evaluation = mt.matrixAddByCol(evaluation,
                                               comp.evaluation)
                state = mt.matrixAddByRow(state, comp.meanPrior)
                sysVar = mt.matrixAddInDiag(sysVar, comp.covPrior)
                self.discount = np.concatenate((self.discount, comp.discount))
                self.componentIndex[i] = (currentIndex,
                                          currentIndex + comp.d - 1)
                currentIndex += comp.d

        # if the model contains the automatic dynamic part, we add
        # them to the builder
        if len(self.automaticComponents) > 0:
            self.automaticEvaluation = None
            for i in self.automaticComponents:
                comp = self.automaticComponents[i]
                comp.updateEvaluation(0, data)
                transition = mt.matrixAddInDiag(transition, comp.transition)
                evaluation = mt.matrixAddByCol(evaluation,
                                               comp.evaluation)
                state = mt.matrixAddByRow(state, comp.meanPrior)
                sysVar = mt.matrixAddInDiag(sysVar, comp.covPrior)
                self.discount = np.concatenate((self.discount, comp.discount))
                self.componentIndex[i] = (currentIndex,
                                          currentIndex + comp.d - 1)
                currentIndex += comp.d

        self.statePrior = state
        self.sysVarPrior = sysVar
        self.noiseVar = np.matrix(noise)
        self.model = baseModel(transition=transition,
                               evaluation=evaluation,
                               noiseVar=np.matrix(noise),
                               sysVar=sysVar,
                               state=state,
                               df=self.initialDegreeFreedom)
        self.model.initializeObservation()

        # compute the renew period
        if self.renewDiscount is None:
            self.renewDiscount = np.min(self.discount)

        if self.renewDiscount < 1.0 - 1e-8:
            self.renewTerm = np.log(0.001 * (1 - self.renewDiscount)) \
                             / np.log(self.renewDiscount)

        self.initialized = True
        self._logger.info('Initialization finished.')

    # Initialize from another builder exported from other dlm class
    def initializeFromBuilder(self, data, exported_builder):
        # Copy the components
        self.staticComponents = deepcopy(exported_builder.staticComponents)
        self.automaticComponents = deepcopy(exported_builder.automaticComponents)
        self.dynamicComponents = deepcopy(exported_builder.dynamicComponents)
        self.componentIndex = deepcopy(exported_builder.componentIndex)
        self.discount = deepcopy(exported_builder.discount)
        self.initialDegreeFreedom = exported_builder.model.df

        # Copy the model states
        self.statePrior = deepcopy(exported_builder.statePrior)
        self.sysVarPrior = deepcopy(exported_builder.sysVarPrior)
        self.noiseVar = deepcopy(exported_builder.noiseVar)
        self.model = deepcopy(exported_builder.model)

        # update the evaluation to the current.
        self.updateEvaluation(step=0, data=data)
        self.model.initializeObservation()

        # compute the renew period
        if self.renewDiscount is None:
            self.renewDiscount = np.min(self.discount)

        if self.renewDiscount < 1.0 - 1e-8:
            self.renewTerm = np.log(0.001 * (1 - self.renewDiscount)) \
                             / np.log(self.renewDiscount)
        
        self.initialized = True
        self._logger.info('Initialization finished.')
            
    # This function allows the model to update the dynamic evaluation vector,
    # so that the model can handle control variables
    # This function should be called only when dynamicComponents is not empty
    # data is used by auto regressor.
    def updateEvaluation(self, step, data):
        """ Update the evaluation matrix of the model to a specific date.
        It loops over all dynamic components and update their evaluation
        matrix and then reconstruct the model evaluation matrix by
        incorporating the new evaluations

        Arges:
            step: the date at which the evaluation matrix is needed.

        """

        # if len(self.dynamicComponents) == 0 and \
        #   len(self.automaticComponents) == 0:
        #    raise NameError('This shall only be used when there' +
        #                    ' are dynamic or automatic components!')

        # update the dynamic evaluation vector
        # We need first update all dynamic components by 1 step
        for i in self.dynamicComponents:
            comp = self.dynamicComponents[i]
            comp.updateEvaluation(step)
            self.model.evaluation[0, self.componentIndex[i][0]:
                                  (self.componentIndex[i][1] + 1)] = comp.evaluation

        for i in self.automaticComponents:
            comp = self.automaticComponents[i]
            comp.updateEvaluation(step, data)
            self.model.evaluation[0, self.componentIndex[i][0]:
                                  (self.componentIndex[i][1] + 1)] = comp.evaluation



import numpy as np


## tools for concatenate matrix ###
class matrixTools:
    @staticmethod
    def matrixAddInDiag(A, B):
        if A is None:
            return np.matrix(B)
        elif B is None:
            return np.matrix(A)
        else:
            (An, Ap) = A.shape
            (Bn, Bp) = B.shape

            newMatrixA = np.concatenate((A, np.matrix(np.zeros((An, Bp)))), 1)
            newMatrixB = np.concatenate((np.matrix(np.zeros((Bn, Ap))), B), 1)
            return np.concatenate((newMatrixA, newMatrixB), 0)


    # A + B = (A; B)
    @staticmethod
    def matrixAddByRow(A, B):
        if A is None:
            return B
        elif B is None:
            return A
        else:
            return np.concatenate((A, B), 0)


    # A + B = (A B)
    @staticmethod
    def matrixAddByCol(A, B):
        if A is None:
            return np.matrix(B)
        elif B is None:
            return np.matrix(A)
        else:
            return np.concatenate((A, B), 1)


    @staticmethod
    def AddTwoVectors(a, b):
        if a is None:
            return np.array(b)
        elif b is None:
            return np.array(a)
        else:
            return np.concatenate((a, b))



import numpy as np
from pydlm.modeler.dynamic import dynamic
import unittest


class testDynamic(unittest.TestCase):


    def setUp(self):
        self.features = np.matrix(np.random.rand(10, 2)).tolist()
        self.newDynamic = dynamic(features=self.features, w=1.0)


    def testInputNumpyMatrix(self):
        dynamic(features=np.random.rand(10, 2), w=1.0)
        pass


    def testInitialization(self):
        self.assertEqual(self.newDynamic.d, 2)
        self.assertEqual(self.newDynamic.n, 10)
        self.assertAlmostEqual(np.abs(np.sum(np.matrix(
            self.newDynamic.features)
                - np.matrix(self.features))), 0)


    def testUpdate(self):
        self.newDynamic.updateEvaluation(3)
        self.assertAlmostEqual(np.abs(
            np.sum(np.array(self.newDynamic.evaluation)
                   - np.array(self.features[3]))), 0)


    def testAppendNewData(self):
        self.newDynamic.appendNewData([[1, 2]])
        self.assertAlmostEqual(np.abs(
            np.sum(np.array(self.newDynamic.features[-1])
                   - np.array([1, 2]))), 0)


    def testPopout(self):
        self.newDynamic.popout(0)
        self.assertAlmostEqual(np.abs(np.sum(
            np.matrix(self.newDynamic.features)
            - np.matrix(self.features[1:]))), 0)


    def testAlter(self):
        self.newDynamic.alter(1, [0, 0])
        self.assertAlmostEqual(self.newDynamic.features[1],
                               [0, 0])


if __name__ == '__main__':
    unittest.main()



from pydlm.modeler.trends import trend
from pydlm.core._dlm import _dlm

import logging
import numpy as np
import unittest


class test_dlm(unittest.TestCase):


    def setUp(self):
        self.data = [0] * 9 + [1] + [0] * 10
        self.dlm1 = _dlm(self.data)
        self.dlm2 = _dlm(self.data)
        self.dlm1.builder + trend(degree=0, discount=1, w=1.0)
        self.dlm2.builder + trend(degree=0, discount=1e-12, w=1.0)
        self.dlm1._initialize()
        self.dlm2._initialize()
        self.dlm1.options.innovationType='whole'
        self.dlm2.options.innovationType='whole'


    def testForwardFilter(self):
        self.dlm1._forwardFilter(start=0, end=19, renew=False)
        self.assertAlmostEqual(np.sum(self.dlm1.result.filteredObs[0:9]), 0)
        self.assertAlmostEqual(self.dlm1.result.filteredObs[9][0, 0], 1.0/11)
        self.assertAlmostEqual(self.dlm1.result.filteredObs[19][0, 0], 1.0/21)

        self.dlm2._forwardFilter(start=0, end=19)
        self.assertAlmostEqual(np.sum(self.dlm2.result.filteredObs[0:9]), 0.0)
        self.assertAlmostEqual(self.dlm2.result.filteredObs[9][0, 0], 1.0)
        self.assertAlmostEqual(self.dlm2.result.filteredObs[19][0, 0], 0.0)


    def testResetModelStatus(self):
        self.dlm1._forwardFilter(start = 0, end = 19, renew = False)
        self.dlm1.result.filteredSteps = (0, 19)
        self.assertAlmostEqual(self.dlm1.builder.model.obs, \
                               self.dlm1.result.filteredObs[19])

        self.dlm1._resetModelStatus()
        self.assertAlmostEqual(np.sum(self.dlm1.builder.model.state \
                                      - self.dlm1.builder.statePrior), 0.0)


    def testSetModelStatus(self):
        self.dlm1._forwardFilter(start = 0, end = 19, renew = False)
        self.dlm1.result.filteredSteps = (0, 19)
        self.assertAlmostEqual(self.dlm1.builder.model.obs, \
                               self.dlm1.result.filteredObs[19])
        self.dlm1._setModelStatus(date = 12)
        self.assertAlmostEqual(self.dlm1.builder.model.obs, \
                               self.dlm1.result.filteredObs[12])


    def testForwaredFilterConsectiveness(self):
        self.dlm1._forwardFilter(start = 0, end = 19, renew = False)
        filtered1 = self.dlm1.result.filteredObs

        self.dlm1._initialize()

        self.dlm1._forwardFilter(start = 0, end = 13)
        self.dlm1.result.filteredSteps = (0, 13)
        self.dlm1._forwardFilter(start = 13, end = 19)
        filtered2 = self.dlm1.result.filteredObs

        self.assertAlmostEqual(np.sum(np.array(filtered1) - np.array(filtered2)), 0.0)


    def testBackwardSmoother(self):
        self.dlm1._forwardFilter(start = 0, end = 19, renew = False)
        self.dlm1.result.filteredSteps = (0, 19)
        self.dlm1._backwardSmoother(start = 19)
        self.assertAlmostEqual(self.dlm1.result.smoothedObs[0][0, 0], 1.0/21)
        self.assertAlmostEqual(self.dlm1.result.smoothedObs[19][0, 0], 1.0/21)

        self.dlm2._forwardFilter(start = 0, end = 19)
        self.dlm2.result.filteredSteps = (0, 19)
        self.dlm2._backwardSmoother(start = 19)
        self.assertAlmostEqual(self.dlm2.result.smoothedObs[0][0, 0], 0.0)
        self.assertAlmostEqual(self.dlm2.result.smoothedObs[19][0, 0], 0.0)
        self.assertAlmostEqual(self.dlm2.result.smoothedObs[9][0, 0], 1.0)


    def testLogger(self):
        assert self.dlm1._logger == logging.getLogger('pydlm')
        assert self.dlm2._logger == logging.getLogger('pydlm')
        assert self.dlm1._logger.getEffectiveLevel() == logging.INFO
        assert self.dlm2._logger.getEffectiveLevel() == logging.INFO

    def testSetAndGetLoggingLevel(self):
        self.dlm1.setLoggingLevel('CRITICAL')
        assert self.dlm1._logger.getEffectiveLevel() == logging.CRITICAL
        assert self.dlm1.getLoggingLevel() == 'CRITICAL'

        self.dlm1.setLoggingLevel('INFO')
        assert self.dlm1._logger.getEffectiveLevel() == logging.INFO
        assert self.dlm1.getLoggingLevel() == 'INFO'


if __name__ == '__main__':
    unittest.main()



"""
===============================================================================

The code for all tuning methods

===============================================================================

"""

from pydlm.core._dlm import _dlm


class _dlmTune(_dlm):
    """ The main class containing all tuning methods.

    Methods:
        _getMSE: obtain the fitting model one-day ahead prediction MSE.
        _getDiscounts: obtain the discounts (for different components).
        _setDiscounts: set discounts for different components.
    """


    # get the mse from the model
    def _getMSE(self):
        
        if not self.initialized:
            raise NameError('need to fit the model first')

        if self.result.filteredSteps[1] == -1:
            raise NameError('need to run forward filter first')

        mse = 0
        for i in range(self.result.filteredSteps[0],
                       self.result.filteredSteps[1] + 1):
            if self.data[i] is not None:
                mse += (self.data[i] - self.result.predictedObs[i]) ** 2

        mse = mse / (self.result.filteredSteps[1] + 1 -
                      self.result.filteredSteps[0])
        return mse[0,0]


    # get the discount from the model
    def _getDiscounts(self):
        
        if not self.initialized:
            raise NameError('need to fit the model before one can'
                            'fetch the discount factors')

        discounts = []
        for comp in self.builder.componentIndex:
            indx = self.builder.componentIndex[comp]
            discounts.append(self.builder.discount[indx[0]])
        return discounts


    # set the model discount, this should never expose to the user
    # change the discount in the component would change the whole model.
    # change those in filter and builder only change the discounts
    # temporarily and will be corrected if we rebuild the model.
    def _setDiscounts(self, discounts, change_component=False):

        if not self.initialized:
            raise NameError('need to fit the model first')

        for i, comp in enumerate(self.builder.componentIndex):
            indx = self.builder.componentIndex[comp]
            self.builder.discount[indx[0]: (indx[1] + 1)] = discounts[i]
            if change_component:
                component = self._fetchComponent(name=comp)
                component.discount = self.builder.discount[indx[0]: (indx[1] + 1)]

        self.Filter.updateDiscount(self.builder.discount)
        self.result.filteredSteps = [0, -1]




import unittest
from pydlm.modeler.longSeason import longSeason


class testLongSeason(unittest.TestCase):


    def setUp(self):
        data = [0, 1, 2, 3, 0, 1, 2, 3, 0, 1, 2, 3]
        self.longSeason = longSeason(data=data, period=4, stay=4, w=1.0)


    def testFeatureMatrix(self):
        trueFeatures = [[1, 0, 0, 0], [1, 0, 0, 0], [1, 0, 0, 0], [1, 0, 0, 0],
                        [0, 1, 0, 0], [0, 1, 0, 0], [0, 1, 0, 0], [0, 1, 0, 0],
                        [0, 0, 1, 0], [0, 0, 1, 0], [0, 0, 1, 0], [0, 0, 1, 0]]
        self.assertEqual(self.longSeason.features, trueFeatures)
        self.assertEqual(self.longSeason.nextState, [3, 0])
        self.assertEqual(self.longSeason.n, 12)


    def testAppendNewData(self):
        trueFeatures = [[1, 0, 0, 0], [1, 0, 0, 0], [1, 0, 0, 0], [1, 0, 0, 0],
                        [0, 1, 0, 0], [0, 1, 0, 0], [0, 1, 0, 0], [0, 1, 0, 0],
                        [0, 0, 1, 0], [0, 0, 1, 0], [0, 0, 1, 0], [0, 0, 1, 0],
                        [0, 0, 0, 1]]
        self.longSeason.appendNewData([4])
        self.assertEqual(self.longSeason.features, trueFeatures)
        self.assertEqual(self.longSeason.nextState, [3, 1])
        self.assertEqual(self.longSeason.n, 13)


    def testPopoutTheLast(self):
        trueFeatures = [[1, 0, 0, 0], [1, 0, 0, 0], [1, 0, 0, 0], [1, 0, 0, 0],
                        [0, 1, 0, 0], [0, 1, 0, 0], [0, 1, 0, 0], [0, 1, 0, 0],
                        [0, 0, 1, 0], [0, 0, 1, 0], [0, 0, 1, 0]]
        self.longSeason.popout(11)
        self.assertEqual(self.longSeason.features, trueFeatures)
        self.assertEqual(self.longSeason.nextState, [2, 3])
        self.assertEqual(self.longSeason.n, 11)


    def testPopoutInMiddel(self):
        trueFeatures = [[1, 0, 0, 0], [1, 0, 0, 0], [1, 0, 0, 0], [1, 0, 0, 0],
                        [0, 1, 0, 0], [0, 1, 0, 0], [0, 1, 0, 0], [0, 1, 0, 0],
                        [0, 0, 1, 0], [0, 0, 1, 0], [0, 0, 1, 0]]
        self.longSeason.popout(1)
        self.assertEqual(self.longSeason.features, trueFeatures)
        self.assertEqual(self.longSeason.nextState, [2, 3])
        self.assertEqual(self.longSeason.n, 11)


    def testUpdateEvaluation(self):
        trueFeatures = [[1, 0, 0, 0], [1, 0, 0, 0], [1, 0, 0, 0], [1, 0, 0, 0],
                        [0, 1, 0, 0], [0, 1, 0, 0], [0, 1, 0, 0], [0, 1, 0, 0],
                        [0, 0, 1, 0], [0, 0, 1, 0], [0, 0, 1, 0], [0, 0, 1, 0],
                        [0, 0, 0, 1]]
        self.longSeason.updateEvaluation(12)
        self.assertEqual(self.longSeason.features, trueFeatures)
        self.assertEqual(self.longSeason.nextState, [3, 1])
        self.assertEqual(self.longSeason.n, 12)


if __name__ == '__main__':
    unittest.main()



import unittest

from pydlm.modeler.trends import trend
from pydlm.dlm import dlm

class testDlmPlot(unittest.TestCase):

    def testPlot(self):
        dlm1 = dlm(range(100)) + trend(1)
        dlm1.fit()
        dlm1.plot()

if __name__ == '__main__':
    unittest.main()



import unittest

from pydlm.modeler.trends import trend
from pydlm.modeler.autoReg import autoReg
from pydlm.access._dlmGet import _dlmGet


class test_dlmGet(unittest.TestCase):


    def setUp(self):
        self.data5 = range(100)
        self.dlm5 = _dlmGet(self.data5)
        self.dlm5.builder + trend(degree=0, discount=1, w=1.0) + \
            autoReg(degree=1, discount=1, w=1.0)
        self.dlm5._initialize()
        self.dlm5.options.innovationType='whole'


    def testGetLatentState(self):
        # for forward filter
        self.dlm5._forwardFilter(start=0, end=99, renew=False)
        self.dlm5.result.filteredSteps = [0, 99]
        filteredTrend = self.dlm5._getLatentState(
            filterType='forwardFilter', name='trend', start=0, end=99)
        diff = 0.0
        for i in range(len(filteredTrend)):
            diff += abs(filteredTrend[i][0] -
                        self.dlm5.result.filteredState[i][0, 0])
        self.assertAlmostEqual(diff, 0)

        # for prediction
        predictedTrend = self.dlm5._getLatentState(
            filterType='predict', name='trend', start=0, end=99)
        diff = 0.0
        for i in range(len(predictedTrend)):
            diff += abs(predictedTrend[i][0] -
                        self.dlm5.result.predictedState[i][0, 0])
        self.assertAlmostEqual(diff, 0)

        # for backward smoother
        self.dlm5._backwardSmoother(start=99)
        self.dlm5.result.smoothedSteps = [0, 99]
        smoothedTrend = self.dlm5._getLatentState(
            filterType='backwardSmoother', name='trend', start=0, end=99)
        diff = 0.0
        for i in range(len(smoothedTrend)):
            diff += abs(smoothedTrend[i][0] -
                        self.dlm5.result.smoothedState[i][0, 0])
        self.assertAlmostEqual(diff, 0)


    def testGetLatentCov(self):
        # for forward filter
        self.dlm5._forwardFilter(start=0, end=99, renew=False)
        self.dlm5.result.filteredSteps = [0, 99]
        filteredTrend = self.dlm5._getLatentCov(
            filterType='forwardFilter', name='trend', start=0, end=99)
        diff = 0.0
        for i in range(len(filteredTrend)):
            diff += abs(filteredTrend[i][0, 0] -
                        self.dlm5.result.filteredCov[i][0, 0])
        self.assertAlmostEqual(diff, 0)

        # for prediction
        predictedTrend = self.dlm5._getLatentCov(
            filterType='predict', name='trend', start=0, end=99)
        diff = 0.0
        for i in range(len(predictedTrend)):
            diff += abs(predictedTrend[i][0, 0] -
                        self.dlm5.result.predictedCov[i][0, 0])
        self.assertAlmostEqual(diff, 0)

        # for backward smoother
        self.dlm5._backwardSmoother(start=99)
        self.dlm5.result.smoothedSteps = [0, 99]
        smoothedTrend = self.dlm5._getLatentCov(
            filterType='backwardSmoother', name='trend', start=0, end=99)
        diff = 0.0
        for i in range(len(smoothedTrend)):
            diff += abs(smoothedTrend[i][0, 0] -
                        self.dlm5.result.smoothedCov[i][0, 0])
        self.assertAlmostEqual(diff, 0)


    def testComponentMean(self):
        self.dlm5._forwardFilter(start=0, end=99, renew=False)
        self.dlm5.result.filteredSteps = [0, 99]
        # for component with forward filter
        arTrend = self.dlm5._getComponentMean(filterType='forwardFilter',
                                              name='ar2', start=0, end=99)
        trueAr = [item[1, 0] for item in self.dlm5.result.filteredState]
        comp = self.dlm5.builder.automaticComponents['ar2']
        for i in range(len(trueAr)):
            comp.updateEvaluation(i, self.data5)
            trueAr[i] = comp.evaluation * trueAr[i]

        diff = 0.0
        for i in range(len(arTrend)):
            diff += abs(arTrend[i] - trueAr[i])
        self.assertAlmostEqual(diff, 0)

        # for component with backward smoother
        self.dlm5._backwardSmoother(start=99)
        self.dlm5.result.smoothedSteps = [0, 99]
        arTrend = self.dlm5._getComponentMean(filterType='backwardSmoother',
                                    name='ar2', start=0, end=99)
        trueAr = [item[1, 0] for item in self.dlm5.result.smoothedState]
        comp = self.dlm5.builder.automaticComponents['ar2']
        for i in range(len(trueAr)):
            comp.updateEvaluation(i, self.data5)
            trueAr[i] = comp.evaluation * trueAr[i]

        diff = 0.0
        for i in range(len(arTrend)):
            diff += abs(arTrend[i] - trueAr[i])
        self.assertAlmostEqual(diff, 0)


    def testComponentVar(self):
        self.dlm5._forwardFilter(start=0, end=99, renew=False)
        self.dlm5.result.filteredSteps = [0, 99]
        # for component with forward filter
        arTrend = self.dlm5._getComponentVar(filterType='forwardFilter',
                                             name='ar2', start=0, end=99)
        trueAr = [item[1, 1] for item in self.dlm5.result.filteredCov]
        comp = self.dlm5.builder.automaticComponents['ar2']
        for i in range(len(trueAr)):
            comp.updateEvaluation(i, self.data5)
            trueAr[i] = comp.evaluation * trueAr[i] * comp.evaluation.T

        diff = 0.0
        for i in range(len(arTrend)):
            diff += abs(arTrend[i] - trueAr[i])
        self.assertAlmostEqual(diff, 0)

        # for component with backward smoother
        self.dlm5._backwardSmoother(start=99)
        self.dlm5.result.smoothedSteps = [0, 99]
        arTrend = self.dlm5._getComponentVar(filterType='backwardSmoother',
                                             name='ar2', start=0, end=99)
        trueAr = [item[1, 1] for item in self.dlm5.result.smoothedCov]
        comp = self.dlm5.builder.automaticComponents['ar2']
        for i in range(len(trueAr)):
            comp.updateEvaluation(i, self.data5)
            trueAr[i] = comp.evaluation * trueAr[i] * comp.evaluation.T

        diff = 0.0
        for i in range(len(arTrend)):
            diff += abs(arTrend[i] - trueAr[i])
        self.assertAlmostEqual(diff, 0)


if __name__ == '__main__':
    unittest.main()



#==================================================================
#
#  This example use the data from Google data science blog post
#  http://www.unofficialgoogledatascience.com/2017/07/fitting-bayesian-structural-time-series.html
#
#  Make sure your working directly is pointing to the folder where
#  `main.py` locates and `data.csv` is in the same folder.
#
#==================================================================

# Read data file
import os
this_dir = os.getcwd()
DATA_PATH = os.path.join(this_dir, "data.csv")
data_file = open(DATA_PATH, 'r')

variables = data_file.readline().strip().split(',')
data_map = {}
for var in variables:
    data_map[var] = []

for line in data_file:
    for i, data_piece in enumerate(line.strip().split(',')):
        data_map[variables[i]].append(float(data_piece))

# Extract and store the data.
time_series = data_map[variables[0]]
features = [[data_map[variables[j]][i] for j in range(1, len(variables)) ]
            for i in range(len(time_series))]

# Plot the raw data
import matplotlib.pyplot as plt
import pydlm.plot.dlmPlot as dlmPlot
dlmPlot.plotData(range(len(time_series)),
                 time_series,
                 showDataPoint=False,
                 label='raw_data')
plt.legend(loc='best', shadow=True)
plt.show()

# Build a simple model
from pydlm import dlm, trend, seasonality

# A linear trend
linear_trend = trend(degree=1, discount=0.95, name='linear_trend', w=10)
# A seasonality
seasonal52 = seasonality(period=52, discount=0.99, name='seasonal52', w=10)

simple_dlm = dlm(time_series) + linear_trend + seasonal52
simple_dlm.fit()

# Plot the fitted results
simple_dlm.turnOff('data points')
simple_dlm.plot()
# Plot each component (attribution)
simple_dlm.turnOff('predict plot')
simple_dlm.turnOff('filtered plot')
simple_dlm.plot('linear_trend')
simple_dlm.plot('seasonal52')
# Plot the prediction give the first 350 weeks and forcast the next 200 weeks.
simple_dlm.plotPredictN(N=200, date=350)
# Plot the prediction give the first 250 weeks and forcast the next 200 weeks.
simple_dlm.plotPredictN(N=200, date=250)

# Build a dynamic regression model
from pydlm import dynamic
regressor10 = dynamic(features=features, discount=1.0, name='regressor10', w=10)
drm = dlm(time_series) + linear_trend + seasonal52 + regressor10
drm.fit()

# Plot the fitted results
drm.turnOff('data points')
drm.plot()
# Plot each component (attribution)
drm.turnOff('predict plot')
drm.turnOff('filtered plot')
drm.plot('linear_trend')
drm.plot('seasonal52')
drm.plot('regressor10')
# Plot the prediction give the first 300 weeks and forcast the next 150 weeks.
drm.plotPredictN(N=150, date=300)
# Plot the prediction give the first 250 weeks and forcast the next 200 weeks.
drm.plotPredictN(N=200, date=250)



import pydlm
import unittest

class testSeasonality(unittest.TestCase):


    def setUp(self):
        self.DEGREE = 7


    def testInitialization(self):
        newSeasonality = pydlm.modeler.seasonality.seasonality(self.DEGREE)
        newSeasonality.checkDimensions()


if __name__ == '__main__':
    unittest.main()
        



"""
===============================================

Base code for running Kalman filter

===============================================

This module implements a Kalman filter that is slightly different fromt the
standard one, following West and Harrison (1999). This Kalman filter accepts
one-dimension discounting factor to adaptively learn the innovation matrix
itself, instead of accepting it from the user. (Although such option is still
provided)

"""
# This code take care of the Kalman filter
import numpy as np
import pydlm.base.tools as tl

# Define the class of Kalman filter which offers a forward filter
# backward smoother and backward sampler for one-step move


class kalmanFilter:
    """ The kalmanFilter class the provide the basic functionalities

    Attributes:
        discount: the discounting factor determining how much information to carry on
        updateInnovation: indicate whether the innovation matrix should be updated.
                          default to True.

    Methods:
        predict: predict one step ahead of the current state
        forwardFilter: one step filter on the model given a new observation
        backwardSmoother: one step backward smooth given the future model and the
                          filtered state and systematic covariance
        backwardSampler: similar to backwardSmoother, using sampling instead of
                         deterministic equations.
        updateDiscount: for updating the discount factors
    """


    def __init__(self, discount=[0.99], \
                 updateInnovation='whole',
                 index=None):
        """ Initializing the kalmanFilter class

        Args:
            discount: the discounting factor, could be a vector
            updateInnovation: the indicator for whether updating innovation matrix

        """

        self.__checkDiscount__(discount)
        self.discount = np.matrix(np.diag(1 / np.sqrt(np.array(discount))))
        self.updateInnovation = updateInnovation
        self.index = index


    def predict(self, model, dealWithMissingEvaluation = False):
        """ Predict the next states of the model by one step

        Args:
            model: the @baseModel class provided all necessary information
            dealWithMissingValue: indicate whether we need to treat the missing value.
                                  it will be turned off when used in forwardFilter as
                                  missing cases have already been address by forwardFilter
        Returns:
            The predicted result is stored in 'model.prediction'

        """
        # check whether evaluation has missing data, if so, we need to take care of it
        if dealWithMissingEvaluation:
            loc = self._modifyTransitionAccordingToMissingValue(model)

        # if the step number == 0, we use result from the model state
        if model.prediction.step == 0:
            model.prediction.state = np.dot(model.transition, model.state)
            model.prediction.obs = np.dot(model.evaluation, model.prediction.state)
            model.prediction.sysVar = np.dot(np.dot(model.transition, model.sysVar),
                                             model.transition.T)

            # update the innovation
            if self.updateInnovation == 'whole':
                self.__updateInnovation__(model)
            elif self.updateInnovation == 'component':
                self.__updateInnovation2__(model)

            # add the innovation to the system variance
            model.prediction.sysVar += model.innovation

            model.prediction.obsVar = np.dot(np.dot(model.evaluation, \
                                                    model.prediction.sysVar), \
                                             model.evaluation.T) + model.noiseVar
            model.prediction.step = 1

        # otherwise, we use previous result to predict next time stamp
        else:
            model.prediction.state = np.dot(model.transition, model.prediction.state)
            model.prediction.obs = np.dot(model.evaluation, model.prediction.state)
            model.prediction.sysVar = np.dot(np.dot(model.transition, \
                                                    model.prediction.sysVar),\
                                             model.transition.T)
            model.prediction.obsVar = np.dot(np.dot(model.evaluation, \
                                                    model.prediction.sysVar), \
                                             model.evaluation.T) + model.noiseVar
            model.prediction.step += 1

        # recover the evaluation and the transition matrix
        if dealWithMissingEvaluation:
            self._recoverTransitionAndEvaluation(model, loc)


    def forwardFilter(self, model, y, dealWithMissingEvaluation = False):
        """ The forwardFilter used to run one step filtering given new data

        Args:
            model: the @baseModel provided the basic information
            y: the newly observed data

        Returns:
            The filtered result is stored in the 'model' replacing the old states

        """
        # check whether evaluation has missing data, if so, we need to take care of it
        if dealWithMissingEvaluation:
            loc = self._modifyTransitionAccordingToMissingValue(model)

        # since we have delt with the missing value, we don't need to double treat it.
        self.predict(model, dealWithMissingEvaluation=False)

        # when y is not a missing data
        if y is not None:

            # first obtain the predicted status
            # we make the prediction step equal to 0 to ensure the prediction
            # is based on the model state and innovation is updated correctlly
            # model.prediction.step = 0
            model.prediction.step = 0

            # the prediction error and the correction matrix
            err = y - model.prediction.obs
            correction = np.dot(model.prediction.sysVar, model.evaluation.T) \
                         / model.prediction.obsVar

            # update new states
            model.df += 1
            lastNoiseVar = model.noiseVar # for updating model.sysVar
            model.noiseVar = model.noiseVar * \
                             (1.0 - 1.0 / model.df + \
                              err * err / model.df / model.prediction.obsVar)

            model.state = model.prediction.state + correction * err

            model.sysVar = model.noiseVar[0, 0] / lastNoiseVar[0, 0] * \
                           (model.prediction.sysVar - np.dot(correction, correction.T) * \
                            model.prediction.obsVar[0, 0])

            model.obs = np.dot(model.evaluation, model.state)
            model.obsVar = np.dot(np.dot(model.evaluation, model.sysVar), \
                                  model.evaluation.T) + model.noiseVar
            # update the innovation using discount
            # model.innovation = model.sysVar * (1 / self.discount - 1)

        # when y is missing, then we update the status by the predicted results
        else:
            # we do not update the model.predict.step because we need to take of the case
            # [5, None, None, None]. In such case, we do not add more innovation, because
            # no information is comming in.
            # This is correct because
            # 1. for the first 'None', the step starts from 0 because '5' appears before
            # 2. for the second 'None', the step starts from 1, but the prediction.state
            #    is correct, because now model.state = model.prediciton.state
            # 3. The last 'None' follows the same

            model.state = model.prediction.state
            model.sysVar = model.prediction.sysVar
            model.obs = model.prediction.obs
            model.obsVar = model.prediction.obsVar

        # recover the evaluation and the transition matrix
        if dealWithMissingEvaluation:
            self._recoverTransitionAndEvaluation(model, loc)


    # The backward smoother for a given unsmoothed states at time t
    # what model should store:
    #      model.state: the last smoothed states (t + 1)
    #      model.sysVar: the last smoothed system variance (t + 1)
    #      model.transition: the transition at time t + 1
    #      model.evaluation: the evaluation vector at time t
    #      model.prediction.sysVar: the predicted system variance for time t + 1
    #      model.prediction.state: the predicted state for time t + 1
    #      rawState: the unsmoothed state at time t
    #      rawSysVar: the unsmoothed system variance at time t
    def backwardSmoother(self, model, rawState, rawSysVar):

        """ The backwardSmoother for one step backward smoothing

        Args:
            model: the @baseModel used for backward smoothing, the model shall store
                 the following information
                 model.state: the last smoothed states (t + 1)
                 model.sysVar: the last smoothed system variance (t + 1)
                 model.transition: the transition at time t + 1
                 model.evaluation: the evaluation vector at time t
                 model.prediction.sysVar: the predicted system variance for time t + 1
                 model.prediction.state: the predicted state for time t + 1
                 rawState: the unsmoothed state at time t
                 rawSysVar: the unsmoothed system variance at time t
            rawState: the filtered state at the current time stamp
            rawSysVar: the filtered systematic covariance at the current time stamp

        Returns:
            The smoothed results are stored in the 'model' replacing the filtered result.
        """

        # check whether evaluation has missing data, if so, we need to take care of it
        # if dealWithMissingEvaluation:
        #    loc = self._modifyTransitionAccordingToMissingValue(model)

        #### use generalized inverse to ensure the computation stability #######

        predSysVarInv = self._gInverse(model.prediction.sysVar)

        ################################################

        backward = np.dot(np.dot(rawSysVar, model.transition.T), predSysVarInv)
        model.state = rawState + np.dot(backward, (model.state - model.prediction.state))
        model.sysVar = rawSysVar + \
                       np.dot(np.dot(backward, \
                                     (model.sysVar - model.prediction.sysVar)), backward.T)
        model.obs = np.dot(model.evaluation, model.state)
        model.obsVar = np.dot(np.dot(model.evaluation, model.sysVar), \
                              model.evaluation.T) + model.noiseVar

        # recover the evaluation and the transition matrix
        #if dealWithMissingEvaluation:
        #    self._recoverTransitionAndEvaluation(model, loc)


    def backwardSampler(self, model, rawState, rawSysVar):
        """ The backwardSampler for one step backward sampling

        Args:
            model: the @baseModel used for backward sampling, the model shall store
                 the following information
                 model.state: the last smoothed states (t + 1)
                 model.sysVar: the last smoothed system variance (t + 1)
                 model.transition: the transition at time t + 1
                 model.evaluation: the evaluation vector at time t
                 model.prediction.sysVar: the predicted system variance for time t + 1
                 model.prediction.state: the predicted state for time t + 1
                 rawState: the unsmoothed state at time t
                 rawSysVar: the unsmoothed system variance at time t
            rawState: the filtered state at the current time stamp
            rawSysVar: the filtered systematic covariance at the current time stamp

        Returns:
            The sampled results are stored in the 'model' replacing the filtered result.
        """
        #### use generalized inverse to ensure the computation stability #######

        predSysVarInv = self._gInverse(model.prediction.sysVar)

        ################################################

        backward = np.dot(np.dot(rawSysVar, model.transition.T), predSysVarInv)
        model.state = rawState + np.dot(backward, (model.state - model.prediction.state))
        model.sysVar = rawSysVar + \
                       np.dot(np.dot(backward, \
                                     (model.sysVar - model.prediction.sysVar)), backward.T)
        model.state = np.matrix(np.random.multivariate_normal(model.state.A1, \
                                                              model.sysVar)).T
        model.obs = np.dot(model.evaluation, model.state)
        model.obsVar = np.dot(np.dot(model.evaluation, model.sysVar), \
                              model.evaluation.T) + model.noiseVar
        model.obs = np.matrix(np.random.multivariate_normal(model.obs.A1, \
                                                              model.obsVar)).T


    # for updating the discounting factor
    def updateDiscount(self, newDiscount):
        """ For updating the discounting factor

        Args:
            newDiscount: the new discount factor
        """

        self.__checkDiscount__(newDiscount)
        self.discount = np.matrix(np.diag(1 / np.sqrt(newDiscount)))


    def __checkDiscount__(self, discount):
        """ Check whether the discount fact is within (0, 1)

        """

        for i in range(len(discount)):
            if discount[i] < 0 or discount[i] > 1:
                raise tl.matrixErrors('discount factor must be between 0 and 1')


    # update the innovation
    def __updateInnovation__(self, model):
        """ update the innovation matrix of the model

        """

        model.innovation = np.dot(np.dot(self.discount, model.prediction.sysVar), \
                                      self.discount) - model.prediction.sysVar


    # update the innovation
    def __updateInnovation2__(self, model):
        """ update the innovation matrix of the model, but only for component
        indepdently. (i.e., only add innovation to block diagonals, not on off
        block diagonals)

        """

        innovation = np.dot(np.dot(self.discount, model.prediction.sysVar), \
                                      self.discount) - model.prediction.sysVar
        model.innovation = np.matrix(np.zeros(innovation.shape))
        for name in self.index:
            indx = self.index[name]
            model.innovation[indx[0]: (indx[1] + 1), indx[0]: (indx[1] + 1)] = \
                    innovation[indx[0]: (indx[1] + 1), indx[0]: (indx[1] + 1)]


    # a generalized inverse of matrix A
    def _gInverse(self, A):
        """ A generalized inverse of matrix A

        """
        return np.linalg.pinv(A)


    def _modifyTransitionAccordingToMissingValue(self, model):
        """ When evaluation contains None value, we modify the corresponding entries
        in the transition to deal with the missing value

        """
        loc = []
        for i in range(model.evaluation.shape[1]):
            if model.evaluation[0, i] is None:
                loc.append(i)
                model.transition[i, i] = 0.0
                model.evaluation[0, i] = 0.0
        return loc


    def _recoverTransitionAndEvaluation(self, model, loc):
        """ We recover the transition and evaluation use the results from
        _modifyTransitionAccordingToMissingValue

        """
        for i in loc:
            model.evaluation[0, i] = None
            model.transition[i, i] = 1.0




"""
===============================================================================

The code for all get methods

===============================================================================

"""
from numpy import dot
from pydlm.core._dlm import _dlm


class _dlmGet(_dlm):
    """ The class containing all get methods for dlm class.

    Methods:
        _getComponent: get the component if it is in dlm
        _getLatentState: get the latent state for a given component
        _getLatentCov: get the latent covariance for a given component
        _getComponentMean: get the mean of a given component
        _getComponentVar: get the variance of a given component
    """


    # function to get the corresponding latent state
    def _getLatentState(self, name, filterType, start, end):
        """ Get the latent states of a given component.

        Args:
            name: the name of the component.
            filterType: the type of the latent states to be returned.
                        could be "forwardFilter", "backwardSmoother" or
                        "predict".
            start: the start date for the latent states to be returned.
            end: the end date to be returned.

        Returns:
            A list of latent states.
        """
        end += 1
        indx = self.builder.componentIndex[name]
        patten = lambda x: x if x is None else x[indx[0]:(indx[1] + 1), 0]

        if filterType == 'forwardFilter':
            return list(map(patten, self.result.filteredState[start:end]))
        elif filterType == 'backwardSmoother':
            return list(map(patten, self.result.smoothedState[start:end]))
        elif filterType == 'predict':
            return list(map(patten, self.result.predictedState[start:end]))
        else:
            raise NameError('Incorrect filter type')


    # function to get the corresponding latent covariance
    def _getLatentCov(self, name, filterType, start, end):
        """ Get the latent covariance of a given component.

        Args:
            name: the name of the component.
            filterType: the type of the latent covariance to be returned.
                        could be "forwardFilter", "backwardSmoother" or
                        "predict".
            start: the start date for the latent covariance to be returned.
            end: the end date to be returned.

        Returns:
            A list of latent covariance.
        """
        end += 1
        indx = self.builder.componentIndex[name]
        patten = lambda x: x if x is None \
                 else x[indx[0]:(indx[1] + 1), indx[0]:(indx[1] + 1)]

        if filterType == 'forwardFilter':
            return list(map(patten, self.result.filteredCov[start:end]))
        elif filterType == 'backwardSmoother':
            return list(map(patten, self.result.smoothedCov[start:end]))
        elif filterType == 'predict':
            return list(map(patten, self.result.predictedCov[start:end]))
        else:
            raise NameError('Incorrect filter type')


    # function to get the component mean
    def _getComponentMean(self, name, filterType, start, end):
        """ Get the mean of a given component.

        Args:
            name: the name of the component.
            filterType: the type of the mean to be returned.
                        could be "forwardFilter", "backwardSmoother" or
                        "predict".
            start: the start date for the mean to be returned.
            end: the end date to be returned.

        Returns:
            A list of mean.
        """
        end += 1
        comp = self._fetchComponent(name)
        componentState = self._getLatentState(name=name,
                                              filterType=filterType,
                                              start=start, end=end)
        result = []
        for k, i in enumerate(range(start, end)):
            if name in self.builder.dynamicComponents:
                comp.updateEvaluation(i)
            elif name in self.builder.automaticComponents:
                comp.updateEvaluation(i, self.padded_data)
            result.append(dot(comp.evaluation,
                              componentState[k]).tolist()[0][0])
        return result


    # function to get the component variance
    def _getComponentVar(self, name, filterType, start, end):
        """ Get the variance of a given component.

        Args:
            name: the name of the component.
            filterType: the type of the variance to be returned.
                        could be "forwardFilter", "backwardSmoother" or
                        "predict".
            start: the start date for the variance to be returned.
            end: the end date to be returned.

        Returns:
            A list of variance.
        """
        end += 1
        comp = self._fetchComponent(name)
        componentCov = self._getLatentCov(name=name,
                                          filterType=filterType,
                                          start=start, end=end)
        result = []
        for k, i in enumerate(range(start, end)):
            if name in self.builder.dynamicComponents:
                comp.updateEvaluation(i)
            elif name in self.builder.automaticComponents:
                comp.updateEvaluation(i, self.padded_data)
            result.append(dot(
                dot(comp.evaluation,
                    componentCov[k]), comp.evaluation.T).tolist()[0][0])
        return result



from copy import deepcopy
from numpy import matrix
from pydlm.predict._dlmPredict import _dlmPredict

class dlmPredictModule(_dlmPredict):
    """ A dlm module containing all prediction methods.
    """


    # One day ahead prediction function
    def predict(self, date=None, featureDict=None):
        """ One day ahead predict based on the current data.

        The predict result is based on all the data before date and predict the
        observation at date + days. 

        The prediction could be on the last day and into the future or in 
        the middle of the time series and ignore the rest. For predicting into
        the future, the new features must be supplied to featureDict. For 
        prediction in the middle, the user can still supply the features which
        will be used priorily. The old features will be used if featureDict is
        None.

        Args:
            date: the index when the prediction based on. Default to the
                  last day.
            featureDict: the feature set for the dynamic Components, in a form
                  of {"component_name": feature}. If the featureDict is not
                  supplied, then the algo reuse those stored in the dynamic
                  components. For dates beyond the last day, featureDict must
                  be supplied.

        Returns:
            A tuple. (Predicted observation, variance of the predicted
            observation)

        """
        # the default prediction date
        if date is None:
            date = self.n - 1

        # check if the data on the date has been filtered
        if date > self.result.filteredSteps[1]:
            raise NameError('Prediction can only be made right' +
                            ' after the filtered date')

        # Clear the existing predictModel before the deepcopy to avoid recurrent
        # recurrent copy which could explode the memory and complexity.
        self._predictModel = None
        self._predictModel = deepcopy(self)
        return self._predictModel._oneDayAheadPredict(date=date,
                                                      featureDict=featureDict)


    def continuePredict(self, featureDict=None):
        """ Continue prediction after the one-day ahead predict.

        If users want to have a multiple day prediction, they can opt to use
        continuePredict after predict with new features contained in
        featureDict. For example,

        >>> # predict 3 days after the last day
        >>> myDLM.predict(featureDict=featureDict_day1)
        >>> myDLM.continuePredict(featureDict=featureDict_day2)
        >>> myDLM.continuePredict(featureDict=featureDict_day3)

        The featureDict acts the same way as in predict().

        Args:
            featureDict: the feature set for the dynamic components, stored
                         in a for of {"component name": vector}. If the set
                         was not supplied, then the algo will re-use the old
                         feature. For days beyond the data, the featureDict
                         for every dynamic component must be provided.

        Returns:
            A tupe. (predicted observation, variance)
        """
        if self._predictModel is None:
            raise NameError('continuePredict has to come after predict.')

        return self._predictModel._continuePredict(featureDict=featureDict)


    # N day ahead prediction
    def predictN(self, N=1, date=None, featureDict=None):
        """ N day ahead prediction based on the current data.

        This function is a convenient wrapper of predict() and
        continuePredict(). If the prediction is into the future, i.e, > n, 
        the featureDict has to contain all feature vectors for multiple days
        for each dynamic component. For example, assume myDLM has a component
        named 'spy' which posseses two dimensions,

        >>> featureDict_3day = {'spy': [[1, 2],[2, 3],[3, 4]]}
        >>> myDLM.predictN(N=3, featureDict=featureDict_3day)

        Args:
            N:    The length of days to predict.
            date: The index when the prediction based on. Default to the
                  last day.
            FeatureDict: The feature set for the dynamic Components, in a form
                  of {"component_name": feature}, where the feature must have
                  N elements of feature vectors. If the featureDict is not
                  supplied, then the algo reuse those stored in the dynamic
                  components. For dates beyond the last day, featureDict must
                  be supplied.

        Returns:
            A tuple of two lists. (Predicted observation, variance of the predicted
            observation)

        """
        if N < 1:
            raise NameError('N has to be greater or equal to 1')
        # Take care if features are numpy matrix
        if featureDict is not None:
            for name in featureDict:
                if isinstance(featureDict[name], matrix):
                    featureDict[name] = featureDict[name].tolist()
        predictedObs = []
        predictedVar = []

        # call predict for the first day
        getSingleDayFeature = lambda f, i: ({k: v[i] for k, v in f.items()}
                                            if f is not None else None)
        # Construct the single day featureDict
        featureDictOneDay = getSingleDayFeature(featureDict, 0)
        (obs, var) = self.predict(date=date, featureDict=featureDictOneDay)
        predictedObs.append(obs)
        predictedVar.append(var)

        # Continue predicting the remaining days
        for i in range(1, N):
            featureDictOneDay = getSingleDayFeature(featureDict, i)
            (obs, var) = self.continuePredict(featureDict=featureDictOneDay)
            predictedObs.append(obs)
            predictedVar.append(var)
        return (self._1DmatrixToArray(predictedObs),
                self._1DmatrixToArray(predictedVar))



import unittest

from pydlm.modeler.trends import trend
from pydlm.modeler.seasonality import seasonality
from pydlm.modeler.autoReg import autoReg
from pydlm.tuner._dlmTune import _dlmTune


class test_dlmTune(unittest.TestCase):


    def setUp(self):
        self.data = [0] * 9 + [1] + [0] * 10
        self.data5 = range(100)
        self.dlm1 = _dlmTune(self.data)
        self.dlm2 = _dlmTune(self.data)
        self.dlm6 = _dlmTune(self.data5)
        self.dlm7 = _dlmTune([0, 1, None, 1, 0, 1, -1])
        self.dlm1.builder + trend(degree=0, discount=1, w=1.0)
        self.dlm2.builder + trend(degree=0, discount=1e-12, w=1.0)
        self.dlm6.builder + trend(degree=0, discount=0.9, w=1.0) + \
            seasonality(period=2, discount=0.8, w=1.0) + \
            autoReg(degree=3, discount=1.0)
        self.dlm7.builder + trend(degree=0, discount=1, w=1.0)
        self.dlm1._initialize()
        self.dlm2._initialize()
        self.dlm6._initialize()
        self.dlm7._initialize()
        self.dlm1.options.innovationType='whole'
        self.dlm2.options.innovationType='whole'
        self.dlm6.options.innovationType='whole'
        self.dlm7.options.innovationType='whole'


    def testComputeMSE(self):
        self.dlm1._forwardFilter(start=0, end=19, renew=False)
        self.dlm1.result.filteredSteps=(0, 19)
        mse1 = self.dlm1._getMSE()
        mse_expect = 0
        for i in range(20):
            mse_expect += (self.dlm1.result.predictedObs[i] -
                            self.data[i]) ** 2
        mse_expect /= 20
        self.assertAlmostEqual(mse1, mse_expect)

        self.dlm2._forwardFilter(start=0, end=19, renew=False)
        self.dlm2.result.filteredSteps=(0, 19)
        mse2 = self.dlm2._getMSE()
        mse_expect = 2.0/20
        self.assertAlmostEqual(mse2, mse_expect)

        # Test missing data
        self.dlm7._forwardFilter(start=0, end=6, renew=False)
        self.dlm7.result.filteredSteps=(0, 6)
        mse3 = self.dlm7._getMSE()
        mse_expect = 0
        for i in range(7):
            if self.dlm7.data[i] is not None:
                mse_expect += (self.dlm7.result.predictedObs[i] -
                               self.dlm7.data[i]) ** 2
        mse_expect /= 7
        self.assertAlmostEqual(mse3, mse_expect)


    def testGetDiscount(self):
        discounts = self.dlm6._getDiscounts()
        self.assertTrue(0.9 in discounts)
        self.assertTrue(0.8 in discounts)
        self.assertTrue(1.0 in discounts)


    def testSetDiscount(self):
        self.dlm6._setDiscounts([0.0, 0.1, 0.2], False)
        self.assertTrue(0.0 in self.dlm6.builder.discount)
        self.assertTrue(0.1 in self.dlm6.builder.discount)
        self.assertTrue(0.2 in self.dlm6.builder.discount)
        self.assertTrue(0.9 not in self.dlm6.builder.discount)
        self.assertTrue(0.8 not in self.dlm6.builder.discount)
        self.assertTrue(1.0 not in self.dlm6.builder.discount)

        self.assertAlmostEqual(self.dlm6.builder.staticComponents['trend'].discount,
                               0.9)

        self.dlm6._setDiscounts([0.0, 0.1, 0.2], True)
        self.assertTrue(self.dlm6.builder.staticComponents['trend'].discount in
                        [0.0, 0.1, 0.2])


if __name__ == '__main__':
    unittest.main()



"""
===========================================================================

The code for autoregressive components

===========================================================================

This code implements the autoregressive component as a sub-class of dynamic.
Different from the dynamic component, the features in the autoReg is generated
from the data, and updated according to the data. All other features are
similar to @dynamic.

"""
import pydlm.base.tools as tl

from .component import component
from numpy import matrix

import logging
import numpy as np


class autoReg(component):
    """ The autoReg class allows user to add an autoregressive component to the dlm.
    This code implements the autoregressive component as a child class of
    component. Different from the dynamic component, the features in the
    autoReg is generated from the data, and updated according to the data.

    The latent states of autoReg are aligned in the order of
    [today - degree, today - degree + 1, ..., today - 2, today - 1]. Thus,
    when fetching the latents from autoReg component, use this order to
    correctly align the coefficients.

    Args:
        data (deprecated): Users get a warning if this argument is used.
        degree: the order of the autoregressive component
        discount: the discount factor
        name: the name of the trend component
        w: the value to set the prior covariance. Default to a diagonal
           matrix with 1e7 on the diagonal.
        padding: either 0 or None. The number to be padded for the first degree
                 days, as no previous data is observed to form the feature
                 matrix
    Examples:
        >>>  # create a auto regression component:
        >>> autoReg8 = autoReg(degree=8, name='autoreg8', discount = 0.99)
        >>>  # change the autoReg8 to have covariance with diagonals are 2 and state 1
        >>> autoReg8.createCovPrior(cov = 2)
        >>> autoReg8.createMeanPrior(mean = 1)

    Attributes:
        d: the degree of autoregressive, i.e., how many days to look back
        data (deprecatd): Users get a warning if this argument is used.
        discount factor: the discounting factor
        name: the name of the component
        padding: either 0 or None. The number to be padded for the first degree
                 days, as no previous data is observed to form the feature
                 matrix
        evaluation: the evaluation matrix for this component
        transition: the transition matrix for this component
        covPrior: the prior guess of the covariance matrix of the latent states
        meanPrior: the prior guess of the latent states

    """


    def __init__(self,
                 data=None,  # DEPRECATED
                 degree=2,
                 discount=0.99,
                 name='ar2',
                 w=100,
                 padding=0):

        if data is not None:
            logging.warning('The data argument in autoReg is deprecated. Please avoid using it.')

        self.componentType = 'autoReg'
        self.d = degree
        self.name = name
        self.discount = np.ones(self.d) * discount
        self.padding = padding

        # Initialize all basic quantities
        self.evaluation = None
        self.transition = None
        self.covPrior = None
        self.meanPrior = None

        # create all basic quantities
        self.createTransition()
        self.createCovPrior(scale=w)
        self.createMeanPrior()

        # record current step in case of lost
        self.step = 0


    def createEvaluation(self, step, data):
        """ The evaluation matrix for auto regressor.

        """
        if step > len(data):
            raise NameError("There is no sufficient data for creating autoregressor.")
        # We pad numbers if the step is too early
        self.evaluation = matrix([[self.padding] * (self.d - step) +
                                  list(data[max(0, (step - self.d)) : step])])


    def createTransition(self):
        """ Create the transition matrix.

        For the auto regressor component, the transition matrix is just the identity matrix

        """
        self.transition = np.matrix(np.eye(self.d))


    def createCovPrior(self, cov = None, scale = 1e6):
        """ Create the prior covariance matrix for the latent states

        """
        if cov is None:
            self.covPrior = np.matrix(np.eye(self.d)) * scale
        else:
            self.covPrior = cov * scale


    def createMeanPrior(self, mean = None, scale = 1):
        """ Create the prior latent state

        """
        if mean is None:
            self.meanPrior = np.matrix(np.zeros((self.d, 1))) * scale
        else:
            self.meanPrior = mean * scale


    def checkDimensions(self):
        """ if user supplies their own covPrior and meanPrior, this can
        be used to check if the dimension matches

        """
        tl.checker.checkVectorDimension(self.meanPrior, self.covPrior)


    def updateEvaluation(self, date, data):
        self.createEvaluation(step=date, data=data)


    def appendNewData(self, data):
        """ AutoReg append new data automatically with the main time series. Nothing
        needs to be done here.

        """
        return



import pydlm
import unittest

class testTrend(unittest.TestCase):


    def setUp(self):
        self.DEGREE = 3


    def testInitialization(self):
        newTrend = pydlm.modeler.trends.trend(self.DEGREE)
        newTrend.checkDimensions()


if __name__ == '__main__':
    unittest.main()



from pydlm.tuner._dlmTune import _dlmTune
from pydlm.tuner.dlmTuner import modelTuner

import logging

class dlmTuneModule(_dlmTune):
    """ A dlm model containing all tuning methods
    """


    def getMSE(self):
        """ Get the one-day ahead prediction mean square error. The mse is
        estimated only for days that has been predicted.

        Returns:
            An numerical value
        """

        return self._getMSE()


    def tune(self, maxit=100):
        """ Automatic tuning of the discounting factors. 

        The method will call the model tuner class to use the default parameters
        to tune the discounting factors and change the discount factor permenantly.
        User needs to refit the model after tuning.
        
        If user wants a more refined tuning and not change any property of the
        existing model, they should opt to use the @modelTuner class.
        """
        simpleTuner = modelTuner()

        if self._logger.isEnabledFor(logging.INFO):
            self.fitForwardFilter()
            self._logger.info(f"The current mse is { str(self.getMSE()) }.")
        
        simpleTuner.tune(untunedDLM=self, maxit=maxit)
        self._setDiscounts(simpleTuner.getDiscounts(), change_component=True)
        
        if self._logger.isEnabledFor(logging.INFO):
            self.fitForwardFilter()
            self._logger.info(f"The new mse is { str(self.getMSE()) }.")



"""
===============================================================================

The code for all predicting methods

===============================================================================

"""

from pydlm.core._dlm import _dlm


class _dlmPredict(_dlm):
    """ The main class containing all prediction methods.

    Methods:
        _oneDayAheadPredict: predict one day a head.
        _continuePredict: continue predicting one day after _oneDayAheadPredict
    """


    # Note the following functions will modify the status of the model, so they
    # shall not be directly call through the main model if this behavior is not
    # desired.
    
    # featureDict contains all the features for prediction.
    # It is a dictionary with key equals to the name of the component and
    # the value as the new feature (a list). The function
    # will first use the features provided in this feature dict, if not
    # found, it will fetch the default feature from the component. If
    # it could not find feature for some component, it returns an error
    # The intermediate result will be stored in result.predictStatus as
    # (start_date, next_pred_date, [all_predicted_values]), which will be
    # used by _continuePredict.
    def _oneDayAheadPredict(self, date, featureDict=None):
        """ One day ahead prediction based on the date and the featureDict.
        The prediction could be on the last day and into the future or in 
        the middle of the time series and ignore the rest. For predicting into
        the future, the new features must be supplied to featureDict. For 
        prediction in the middle, the user can still supply the features which
        will be used priorily. The old features will be used if featureDict is
        None.

        Args:
            date: the prediction starts (based on the observation before and
                  on this date)
            featureDict: the new feature value for some dynamic components.
                         must be specified in a form of {component_name: value}
                         if the feature for some dynamic component is not
                         supplied. The algorithm will use the features from
                         the old data. (which means if the prediction is out
                         of sample, then all dynamic component must be provided
                         with the new feature value)

        Returns:
            A tuple of (predicted_mean, predicted_variance)
        """
        if date > self.n - 1:
            raise NameError('The date is beyond the data range.')

        # get the correct status of the model
        self._setModelStatus(date=date)
        self._constructEvaluationForPrediction(
            date=date + 1,
            featureDict=featureDict,
            padded_data=self.padded_data[:(date + 1)])
        
        # initialize the prediction status
        self.builder.model.prediction.step = 0

        # start predicting
        self.Filter.predict(self.builder.model)

        predictedObs = self.builder.model.prediction.obs
        predictedObsVar = self.builder.model.prediction.obsVar
        self.result.predictStatus = [
            date,                   # start_date
            date + 1,               # current_date
            [predictedObs[0, 0]]    # all historical predictions
        ]

        return (predictedObs, predictedObsVar)


    def _continuePredict(self, featureDict=None):
        """ Continue predicting one day after _oneDayAheadPredict or
        after _continuePredict. After using
        _oneDayAheadPredict, the user can continue predicting by using
        _continuePredict. The featureDict act the same as in
        _oneDayAheadPredict.

        Args:
            featureDict: the new feature value for some dynamic components.
                         see @_oneDayAheadPredict

        Returns:
            A tuple of (predicted_mean, predicted_variance)
        """
        if self.result.predictStatus is None:
            raise NameError('_continoousPredict can only be used after ' +
                            '_oneDayAheadPredict')
        startDate = self.result.predictStatus[0]
        currentDate = self.result.predictStatus[1]

        self._constructEvaluationForPrediction(
            date=currentDate + 1,
            featureDict=featureDict,
            padded_data=self.padded_data[:(startDate + 1)] +
                        self.result.predictStatus[2])
        self.Filter.predict(self.builder.model)
        predictedObs = self.builder.model.prediction.obs
        predictedObsVar = self.builder.model.prediction.obsVar
        self.result.predictStatus[1] += 1
        self.result.predictStatus[2].append(predictedObs[0, 0])
        return (predictedObs, predictedObsVar)


    # This function will modify the status of the object, use with caution.
    def _constructEvaluationForPrediction(self,
                                          date,
                                          featureDict=None,
                                          padded_data=None):
        """ Construct the evaluation matrix based on date and featureDict.

        Used for prediction. Features provided in the featureDict will be used
        preferrably. If the feature is not found in featureDict, the algorithm
        will seek it based on the old data and the date.

        Args:
            featureDict: a dictionary containing {dynamic_component_name: value}
                         for update the feature for the corresponding component.
            date: if a dynamic component name is not found in featureDict, the
                  algorithm is using its old feature on the given date.
            padded_data: is the mix of the raw data and the predicted data. It is
                  used by auto regressor.

        """
        # New features are provided. Update dynamic componnet.
        # We distribute the featureDict back to dynamicComponents. If the date is
        # out of bound, we append the feature to the feature set. If the date is
        # within range, we replace the old feature with the new feature.
        if featureDict is not None:
            for name in featureDict:
                if name in self.builder.dynamicComponents:
                    comp = self.builder.dynamicComponents[name]                    
                    # the date is within range
                    if date < comp.n:
                        comp.features[date] = featureDict[name]
                        comp.n += 1
                    elif date < comp.n + 1:
                        comp.features.append(featureDict[name])
                        comp.n += 1
                    else:
                        raise NameError("Feature is missing between the last predicted " +
                                        "day and the new day")
                        
        self.builder.updateEvaluation(date, padded_data)



from setuptools import setup, find_packages

from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name = 'pydlm',
    author = 'Xiangyu Wang',
    author_email = 'wwrechard@gmail.com',
    description = ('A python library for the Bayesian dynamic linear ' +
      'model for time series modeling'),
    long_description=long_description,
    long_description_content_type='text/markdown',
    license = 'BSD',
    keywords = 'dlm bayes bayesian kalman filter smoothing dynamic model',
    url = 'https://github.com/wwrechard/pydlm',
    packages = find_packages(),
    zip_safe= False,
    classifiers = [
      'Development Status :: 3 - Alpha',
      'Intended Audience :: Science/Research',
      'License :: OSI Approved :: BSD License',
      'Operating System :: OS Independent',
      'Programming Language :: Python',
      'Programming Language :: Python :: 3',
      'Topic :: Scientific/Engineering :: Artificial Intelligence'
    ],
    include_package_data = False,
    install_requires = [
      'numpy',
      'matplotlib',
    ],
    tests_require = [
    ],
    extras_require = {
        'docs': [
          'Sphinx',
        ],
    },
)



import numpy as np
import unittest

from pydlm.modeler.trends import trend
from pydlm.modeler.seasonality import seasonality
from pydlm.modeler.builder import builder
from pydlm.base.kalmanFilter import kalmanFilter


class testKalmanFilter(unittest.TestCase):


    def setUp(self):
        self.kf1 = kalmanFilter(discount=[1])
        self.kf0 = kalmanFilter(discount=[1e-10])
        self.kf11 = kalmanFilter(discount=[1, 1])
        self.trend0 = trend(degree=0, discount=1, w=1.0)
        self.trend0_90 = trend(degree=0, discount=0.9, w=1.0)
        self.trend0_98 = trend(degree=0, discount=0.98, w=1.0, name='a')
        self.trend1 = trend(degree=1, discount=1, w=1.0)


    def testForwardFilter(self):
        dlm = builder()
        dlm.add(self.trend0)
        dlm.initialize()
        self.kf1.predict(dlm.model)
        self.assertAlmostEqual(dlm.model.prediction.obs, 0)

        # the prior on the mean is zero, but observe 1, with
        # discount = 1, one should expect the filterd mean to be 0.5
        self.kf1.forwardFilter(dlm.model, 1)
        self.assertAlmostEqual(dlm.model.obs, 0.5)
        self.assertAlmostEqual(dlm.model.prediction.obs, 0)
        self.assertAlmostEqual(dlm.model.sysVar, 0.375)

        self.kf1.predict(dlm.model)
        self.assertAlmostEqual(dlm.model.obs, 0.5)
        self.assertAlmostEqual(dlm.model.prediction.obs, 0.5)

        dlm.initialize()
        self.kf0.predict(dlm.model)
        self.assertAlmostEqual(dlm.model.prediction.obs, 0)

        # the prior on the mean is zero, but observe 1, with discount = 0
        # one should expect the filtered mean close to 1
        self.kf0.forwardFilter(dlm.model, 1)
        self.assertAlmostEqual(dlm.model.obs[0, 0], 1)
        self.assertAlmostEqual(dlm.model.prediction.obs[0, 0], 0)
        self.assertAlmostEqual(dlm.model.sysVar[0, 0], 0.5)

        self.kf0.predict(dlm.model)
        self.assertAlmostEqual(dlm.model.obs[0, 0], 1)
        self.assertAlmostEqual(dlm.model.prediction.obs[0, 0], 1)


    def testForwardFilterMultiDim(self):
        dlm = builder()
        dlm.add(seasonality(period=2, discount=1, w=1.0))
        dlm.initialize()

        self.kf11.forwardFilter(dlm.model, 1)
        self.assertAlmostEqual(dlm.model.state[0][0, 0], 0.33333333333)
        self.assertAlmostEqual(dlm.model.state[1][0, 0], -0.33333333333)

        self.kf11.forwardFilter(dlm.model, -1)
        self.assertAlmostEqual(dlm.model.state[0][0, 0], -0.5)
        self.assertAlmostEqual(dlm.model.state[1][0, 0], 0.5)


    def testBackwardSmoother(self):
        dlm = builder()
        dlm.add(self.trend0)
        dlm.initialize()

        # with mean being 0 and observe 1 and 0 consectively, one shall
        # expect the smoothed mean at 1 will be 1/3, for discount = 1
        self.kf1.forwardFilter(dlm.model, 1)
        self.kf1.forwardFilter(dlm.model, 0)
        self.kf1.backwardSmoother(dlm.model, \
                                  np.matrix([[0.5]]), \
                                  np.matrix([[0.375]]))
        self.assertAlmostEqual(dlm.model.obs[0, 0], 1.0/3)
        self.assertAlmostEqual(dlm.model.sysVar[0, 0], 0.18518519)


    # second order trend with discount = 1. The smoothed result should be
    # equal to a direct fit on the three data points, 0, 1, -1. Thus, the
    # smoothed observation should be 0.0
    def testBackwardSmootherMultiDim(self):
        dlm = builder()
        dlm.add(self.trend1)
        dlm.initialize()

        self.kf11.forwardFilter(dlm.model, 1)
        state1 = dlm.model.state
        cov1 = dlm.model.sysVar

        self.kf11.forwardFilter(dlm.model, -1)
        self.kf11.backwardSmoother(dlm.model, \
                                   rawState = state1, \
                                   rawSysVar = cov1)

        self.assertAlmostEqual(dlm.model.obs[0, 0], 0.0)


    def testMissingData(self):
        dlm = builder()
        dlm.add(self.trend0)
        dlm.initialize()

        self.kf0.forwardFilter(dlm.model, 1)
        self.assertAlmostEqual(dlm.model.obs[0, 0], 1.0)
        self.assertAlmostEqual(dlm.model.obsVar[0, 0], 1.0)

        self.kf0.forwardFilter(dlm.model, None)
        self.assertAlmostEqual(dlm.model.obs[0, 0], 1.0)
        self.assertAlmostEqual(dlm.model.obsVar[0, 0]/1e10, 0.5)

        self.kf0.forwardFilter(dlm.model, None)
        self.assertAlmostEqual(dlm.model.obs[0, 0], 1.0)
        self.assertAlmostEqual(dlm.model.obsVar[0, 0]/1e10, 0.5)

        self.kf0.forwardFilter(dlm.model, 0)
        self.assertAlmostEqual(dlm.model.obs[0, 0], 0.0)


    def testMissingEvaluation(self):
        dlm = builder()
        dlm.add(self.trend0)
        dlm.initialize()

        dlm.model.evaluation = np.matrix([[None]])
        self.kf1.forwardFilter(dlm.model, 1.0, dealWithMissingEvaluation = True)
        self.assertAlmostEqual(dlm.model.obs, 0.0)
        self.assertAlmostEqual(dlm.model.transition, 1.0)


    def testEvolveMode(self):
        dlm = builder()
        dlm.add(self.trend0_90)
        dlm.add(self.trend0_98)
        dlm.initialize()

        kf2 = kalmanFilter(discount=[0.9, 0.98],
                           updateInnovation='component',
                           index=dlm.componentIndex)
        kf2.forwardFilter(dlm.model, 1.0)
        self.assertAlmostEqual(dlm.model.innovation[0, 1], 0.0)
        self.assertAlmostEqual(dlm.model.innovation[1, 0], 0.0)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
"""
===============================================================================

The code for the class dlm

===============================================================================

This is the main class of the Bayeisan dynamic linear model.
It provides the modeling, filtering, forecasting and smoothing function
 of a dlm. The dlm use the @builder to construct the @baseModel based
 on user supplied @components and then run @kalmanFilter to filter the result.

Example:
>>> # randomly generate fake data on 1000 days
>>> import numpy as np
>>> data = np.random.random((1, 1000))

>>> # construct the dlm of a linear trend and a 7-day seasonality
>>> from pydlm import dlm, trend, seasonality
>>> myDlm = dlm(data) + trend(degree = 2, 0.98) + seasonality(period = 7, 0.98)

>>> # filter the result
>>> myDlm.fitForwardFilter()

>>> # extract the filtered result
>>> myDlm.getFilteredObs()

"""
# This is the major class for fitting time series data using the
# dynamic linear model. dlm is a subclass of builder, with adding the
# Kalman filter functionality for filtering the data

from copy import deepcopy
from numpy import matrix
from pydlm.predict.dlmPredictMod import dlmPredictModule
from pydlm.access.dlmAccessMod import dlmAccessModule
from pydlm.tuner.dlmTuneMod import dlmTuneModule
from pydlm.plot.dlmPlotMod import dlmPlotModule


class dlm(dlmPlotModule, dlmPredictModule, dlmAccessModule, dlmTuneModule):
    """ The main class of the dynamic linear model.


    This is the main class of the Bayeisan dynamic linear model.
    It provides the modeling, filtering, forecasting and smoothing
    function of a dlm.
    The dlm use the @builder to construct the @baseModel based on user supplied
    @components and then run @kalmanFilter to filter the result.

    Example 1:
        >>> # randomly generate fake data on 1000 days
        >>> import numpy as np
        >>> data = np.random.random((1, 1000))
        >>> # construct the dlm of a linear trend and a 7-day seasonality
        >>> myDlm = dlm(data) + trend(degree = 2, 0.98) + seasonality(period = 7, 0.98)
        >>> # filter the result
        >>> myDlm.fitForwardFilter()
        >>> # extract the filtered result
        >>> myDlm.getFilteredObs()

    Example 2 (fit a linear regression):
        >>> from pydlm import dynamic
        >>> data = np.random.random((1, 100))
        >>> mydlm = dlm(data) + trend(degree=1, 0.98, name='a') +
                        dynamic(features=[[i] for i in range(100)], 1, name='b')
        >>> mydlm.fit()
        >>> coef_a = mydlm.getLatentState('a')
        >>> coef_b = mydlm.getLatentState('b')

    Attributes:
       data: a list of doubles of the raw time series data.
             It could be either the python's built-in list of
             doubles or numpy 1d array.

    """


    # define the basic members
    # initialize the result
    def __init__(self, data, **options):
        super(dlm, self).__init__(data, **options)

        # This model is used for prediction. Prediction functions
        # will change the model status to forecast at a particular
        # date. Using a copied model will help the main model from
        # being changed and behaving abnormally.
        self._predictModel = None


    def exportModel(self):
        """ Export the dlm builder. Currently the method only support dlm without
        dynamic components.

        """
        if length(self.builder.dynamicComponents) > 0:
            raise ValueError('Cannot export dlm builder with dynamic components.')

        if not self.initialized:
            raise ValueError('Cannot export dlm before the model was initilized.')

        return deepcopy(self.builder)


    def buildFromModel(self, model):
        """ Construct the dlm with exported model from other DLM with status.

        Args:
            model: The exported model from other dlm. Must be the return from
                   dlm.exportModel()

        """
        self._initializeFromBuilder(exported_builder=model)

# ===================== modeling components =====================


    # add component
    def add(self, component):
        """ Add new modeling component to the dlm.

        Currently support: trend, seasonality, autoregression
        and dynamic component.

        Args:
            component: the modeling component, could be either one
                       of the following:\n
                       trend, seasonality, dynamic, autoReg.

        Returns:
            A dlm object with added component.

        """
        self.__add__(component)


    def __add__(self, component):
        self.builder.__add__(component)
        self.initialized = False
        return self


    # list all components
    def ls(self):
        """ List out all existing components

        """
        self.builder.ls()


    # delete one component
    def delete(self, name):
        """ Delete model component by its name

        Args:
            name: the name of the component.

        """
        self.builder.delete(name)
        self.initialized = False

# ========================== model training component =======================


    def fitForwardFilter(self, useRollingWindow=False, windowLength=3):
        """ Fit forward filter on the available data.

        User can choose use rolling windowFront
        or not. If user choose not to use the rolling window,
        then the filtering will be based on all the previous data.
        If rolling window is used, then the filtering for a particular
        date will only consider previous dates that are
        within the rolling window length.

        Args:
            useRollingWindow: indicate whether rolling window should be used.
            windowLength: the length of the rolling window if used.

        """
        # check if the feature size matches the data size
        self._checkFeatureSize()

        # see if the model has been initialized
        if not self.initialized:
            self._initialize()

        self._logger.info('Starting forward filtering...')
        if not useRollingWindow:
            # we start from the last step of previous filtering
            if self.result.filteredType == 'non-rolling':
                start = self.result.filteredSteps[1] + 1
            else:
                start = 0
                # because we refit the forward filter, we need to reset the
                # backward smoother as well.
                self.result.smoothedSteps = [0, -1]

            # determine whether renew should be used
            self._forwardFilter(start=start,
                                end=self.n - 1,
                                renew=self.options.stable)
            self.result.filteredType = 'non-rolling'
        else:
            if self.result.filteredType == 'rolling':
                windowFront = self.result.filteredSteps[1] + 1
            else:
                windowFront = 0
                # because we refit the forward filter, we need to reset the
                # backward smoother as well.
                self.result.smoothedSteps = [0, -1]

            self.result.filteredType = 'rolling'
            # if end is still within (0, windowLength - 1), we should run the
            # usual ff from
            if windowFront < windowLength:
                self._forwardFilter(start=self.result.filteredSteps[1] + 1,
                                    end=min(windowLength - 1, self.n - 1))

            # for the remaining date, we use a rolling window
            for today in range(max(windowFront, windowLength), self.n):
                self._forwardFilter(start=today - windowLength + 1,
                                    end=today,
                                    save=today,
                                    ForgetPrevious=True)

        self.result.filteredSteps = [0, self.n - 1]
        self.turnOn('filtered plot')
        self.turnOn('predict plot')

        self._logger.info('Forward filtering completed.')


    def fitBackwardSmoother(self, backLength=None):
        """ Fit backward smoothing on the data. Starting from the last observed date.

        Args:
            backLength: integer, indicating how many days the backward smoother
            should go, starting from the last date.

        """

        # see if the model has been initialized
        if not self.initialized:
            raise ValueError('Backward Smoother has to be run after' +
                            ' forward filter')

        if self.result.filteredSteps[1] != self.n - 1:
            raise ValueError('Forward Fiter needs to run on full data before' +
                            'using backward Smoother')

        # default value for backLength
        if backLength is None:
            backLength = self.n

        self._logger.info('Starting backward smoothing...')
        # if the smoothed dates has already been done, we do nothing
        if self.result.smoothedSteps[1] == self.n - 1 and \
           self.result.smoothedSteps[0] <= self.n - 1 - backLength + 1:
            return None

        # if the smoothed dates start from n - 1, we just need to continue
        elif self.result.smoothedSteps[1] == self.n - 1:
            self._backwardSmoother(start=self.result.smoothedSteps[0] - 1,
                                   days=backLength)

        # if the smoothed dates are even earlier,
        # we need to start from the beginning
        elif self.result.smoothedSteps[1] < self.n - 1:
            self._backwardSmoother(start=self.n - 1, days=backLength)

        self.result.smoothedSteps = [self.n - backLength, self.n - 1]
        self.turnOn('smoothed plot')

        self._logger.info('Backward smoothing completed.')


    def fit(self):
        """ An easy caller for fitting both the forward filter and backward smoother.

        """
        self.fitForwardFilter()
        self.fitBackwardSmoother()


# ======================= data appending, popping and altering ===============


    # Append new data or features to the dlm
    def append(self, data, component='main'):
        """ Append the new data to the main data or the components (new feature data)

        Args:
            data: the new data
            component: the name of which the new data to be added to.\n
                       'main': the main time series data\n
                       other omponent name: add new feature data to other
                       component.

        """
        # initialize the model to ease the modification
        if not self.initialized:
            self._initialize()
            
        # if we are adding new data to the time series
        if component == 'main':
            # add the data to the self.data
            self.data.extend(list(data))

            # update the length
            self.n += len(data)
            self.result._appendResult(len(data))

            # update the automatic components as well
            for component in self.builder.automaticComponents:
                comp = self.builder.automaticComponents[component]
                comp.appendNewData(data)

            # give a warning to remind to append dynamic components
            if len(self.builder.dynamicComponents) > 0:
                self._logger.warning('Dynamic components need to manually append new data '
                                     'when main data has been appended.')

        # if we are adding new data to the features of dynamic components
        elif component in self.builder.dynamicComponents:
            comp = self.builder.dynamicComponents[component]
            comp.appendNewData(data)

        else:
            raise ValueError('Such dynamic component does not exist.')


    # pop the data of a specific date out
    def popout(self, date):
        """ Pop out the data for a given date

        Args:
            date: the index indicates which date to be popped out.

        """
        if date < 0 or date > self.n - 1:
            raise NameError('The date should be between 0 and ' +
                            str(self.n - 1))

        # initialize the model to ease the modification
        if not self.initialized:
            self._initialize()

        # pop out the data at date
        self.data.pop(date)
        self.n -= 1

        # pop out the feature at date
        for name in self.builder.dynamicComponents:
            comp = self.builder.dynamicComponents[name]
            comp.popout(date)

        # pop out the results at date
        self.result._popout(date)

        # update the filtered and the smoothed steps
        self.result.filteredSteps[1] = date - 1
        self.result.smoothedSteps[1] = date - 1

        if self.result.filteredSteps[0] > self.result.filteredSteps[1]:
            self.result.filteredSteps = [0, -1]
            self.result.smoothedSteps = [0, -1]

        elif self.result.smoothedSteps[0] > self.result.smoothedSteps[1]:
            self.result.smoothedSteps = [0, -1]


    # alter the data of a specific days
    def alter(self, date, data, component='main'):
        """ To alter the data for a specific date and a specific component.

        Args:
            date: the date of the altering data
            data: the new data. data must be a numeric value for main time
                  series and must be a list of numerical values for dynamic
                  components.
            component: the component for which the new data need to be
                       supplied to.\n
                       'main': the main time series data\n
                       other component name: other component feature data

        """
        if date < 0 or date > self.n - 1:
            raise NameError('The date should be between 0 and ' +
                            str(self.n - 1))

        # initialize the model to ease the modification
        if not self.initialized:
            self._initialize()

        # to alter the data for the observed chain
        if component == 'main':
            self.data[date] = data

            # we also automatically alter all the automatic components
            for component in self.builder.automaticComponents:
                comp = self.builder.automaticComponents[component]
                comp.alter(date, data)

        # to alter the feature of a component
        elif component in self.builder.dynamicComponents:
            comp = self.builder.dynamicComponents[component]
            comp.alter(date, data)

        else:
            raise NameError('Such dynamic component does not exist.')

        # update the filtered and the smoothed steps
        self.result.filteredSteps[1] = date - 1
        self.result.smoothedSteps[1] = date - 1

        if self.result.filteredSteps[0] > self.result.filteredSteps[1]:
            self.result.filteredSteps = [0, -1]
            self.result.smoothedSteps = [0, -1]

        elif self.result.smoothedSteps[0] > self.result.smoothedSteps[1]:
            self.result.smoothedSteps = [0, -1]


    # ignore the data of a given date
    def ignore(self, date):
        """ Ignore the data for a specific day. treat it as missing data

        Args:
            date: the date to ignore.

        """
        if date < 0 or date > self.n - 1:
            raise NameError('The date should be between 0 and ' +
                            str(self.n - 1))

        self.alter(date=date, data=None, component='main')

# ================================ control options =========================


    def showOptions(self):
        """ Print out all the option values

        """
        allItems = vars(self.options)
        for item in allItems:
            print(item + ': ' + str(allItems[item]))


    def stableMode(self, use=True):
        """ Turn on the stable mode, i.e., using the renewal strategy.

            Indicate whether the renew strategy should be used to add numerical
            stability. When the filter goes over certain steps,
            the information contribution of the previous data has decayed
            to minimum. In the stable mode, We then ignore those days and
            refit the time series starting from current - renewTerm, where
            renewTerm is computed according to the discount. Thus,
            the effective sample size of the dlm is twice
            renewTerm. When discount = 1, there will be no renewTerm,
            since all the information will be passed along.
        """
        # if option changes, reset everything
        if self.options.stable != use:
            self.initialized = False

        if use is True:
            self.options.stable = True
        elif use is False:
            self.options.stable = False
        else:
            raise ValueError('Incorrect option input')


    def evolveMode(self, evoType='dependent'):
        """ Control whether different component evolve indpendently. If true,
        then the innovation will only be added on each component but not the
        correlation between the components, so that for component with discount
        equals to 1, the smoothed results will always be constant.

        Args:
            evoType: If set to 'independent', then each component will evolve
                     independently. If set to 'dependent', then the components
                     will proceed jointly. Default to 'independent'. Switch to
                     'dependent' if efficiency is a concern.

        Returns:
            a dlm object (for chaining purpose)
        """
        # if option changes, reset everything
        if (self.options.innovationType == 'whole' and
            evoType == 'independent') or \
           (self.options.innovationType == 'component' and
            evoType == 'dependent'):
            self.initialized = False

        if evoType == 'independent':
            self.options.innovationType = 'component'
        elif evoType == 'dependent':
            self.options.innovationType = 'whole'
        else:
            raise NameError('Incorrect option input')

        # for chaining
        return self


    def noisePrior(self, prior=0):
        """ To set the prior for the observational noise. Calling with empty
        argument will enable the auto noise intializer (currently, the min of 1
        and the variance of time series).

        Args:
            prior: the prior of the observational noise.

        Returns:
            A dlm object (for chaining purpose)
        """
        if prior > 0:
            self.options.noise=prior
            self.initialized = False
        else:
            self.options.useAutoNoise = True
            self.initialized = False            

        # for chaining
        return self



from pydlm.core._dlm import _dlm

class dlmPlotModule(_dlm):
    """ A dlm module containing all plot methods
    """
    def __init__(self, data, **options):
        super(dlmPlotModule, self).__init__(data, **options)

        # indicate whether the plot modules has been loaded.
        # We add this flag, since we only import plot module
        # when they are called to avoid any error due to
        # plot that blocks using this package. This flag can
        # help without doing import-check (expensive) every time
        # when plot function is called.
        self.plotLibLoaded = False


    def turnOn(self, switch):
        """ "turn on" Operation for the dlm plotting options.

        Args:
            switch: The key word to switch on. \n
                    'filtered plot', 'filter' to plot filtered results\n
                    'smoothed plot', 'smooth' to plot smoothed results\n
                    'predict plot', 'predict', to plot one-step ahead results\n
                    'confidence interval', 'confierence', 'CI' to plot CI's\n
                    'data points', 'data', 'data point' to plot original data\n
                    'multiple', 'separate' to plot results in separate
                    figures\n
                    'fitted dots', 'fitted' to plot fitted results with dots
        """
        if switch in set(['filtered plot', 'filter',
                          'filtered results', 'filtering']):
            self.options.plotFilteredData = True
        elif switch in set(['smoothed plot', 'smooth',
                            'smoothed results', 'smoothing']):
            self.options.plotSmoothedData = True
        elif switch in set(['predict plot', 'predict',
                            'predicted results', 'prediction']):
            self.options.plotPredictedData = True
        elif switch in set(['confidence interval', 'confidence',
                            'interval', 'CI', 'ci']):
            self.options.showConfidenceInterval = True
        elif switch in set(['data points', 'data point', 'points', 'data']):
            self.options.showDataPoint = True
        elif switch in set(['multiple', 'multiple plots',
                            'separate plots', 'separate']):
            self.options.separatePlot = True
        elif switch in set(['fitted dots', 'fitted results',
                            'fitted data', 'fitted']):
            self.options.showFittedPoint = True
        else:
            raise NameError('no such options')


    def turnOff(self, switch):
        """ "turn off" Operation for the dlm plotting options.

        Args:
            switch: The key word to switch off. \n
                    'filtered plot', 'filter' to not plot filtered results\n
                    'smoothed plot', 'smooth' to not plot smoothed results\n
                    'predict plot', 'predict', to not plot one-step ahead
                    results\n
                    'confidence interval', 'confierence', 'CI'
                    to not plot CI's\n
                    'data points', 'data', 'data point' to not
                    plot original data\n
                    'multiple', 'separate' to not plot results
                    in separate figures\n
                    'fitted dots', 'fitted' to not plot fitted
                    results with dots

        """
        if switch in set(['filtered plot', 'filter', 'filtered results',
                          'filtering']):
            self.options.plotFilteredData = False
        elif switch in set(['smoothed plot', 'smooth', 'smoothed results',
                            'smoothing']):
            self.options.plotSmoothedData = False
        elif switch in set(['predict plot', 'predict', 'predicted results',
                            'prediction']):
            self.options.plotPredictedData = False
        elif switch in set(['confidence interval', 'confidence', 'interval',
                            'CI', 'ci']):
            self.options.showConfidenceInterval = False
        elif switch in set(['data points', 'data point', 'points', 'data']):
            self.options.showDataPoint = False
        elif switch in set(['multiple', 'multiple plots', 'separate plots',
                            'separate']):
            self.options.separatePlot = False
        elif switch in set(['fitted dots', 'fitted results', 'fitted data',
                            'fitted']):
            self.options.showFittedPoint = False
        else:
            raise NameError('no such options')


    def setColor(self, switch, color):
        """ "set" Operation for the dlm plotting colors

        Args:
            switch: key word. Controls over
                    filtered/smoothed/predicted results,
            color: the color for the corresponding keyword.
        """
        if switch in set(['filtered plot', 'filter', 'filtered results',
                          'filtering']):
            self.options.filteredColor = color
        elif switch in set(['smoothed plot', 'smooth', 'smoothed results',
                            'smoothing']):
            self.options.smoothedColor = color
        elif switch in set(['predict plot', 'predict', 'predicted results',
                            'prediction']):
            self.options.predictedColor = color
        elif switch in set(['data points', 'data point', 'points', 'data']):
            self.options.dataColor = color
        else:
            raise NameError('no such options')


    def setConfidence(self, p=0.95):
        """ Set the confidence interval for the plot

        """
        assert p >= 0 and p <= 1
        self.options.confidence = p


    def setIntervalType(self, intervalType):
        """ Set the confidence interval type

        """
        if intervalType == 'ribbon' or intervalType == 'line':
            self.options.intervalType = intervalType
        else:
            raise NameError('No such type for confidence interval.')


    def resetPlotOptions(self):
        """ Reset the plotting option for the dlm class

        """
        self.options.plotOriginalData = True
        self.options.plotFilteredData = True
        self.options.plotSmoothedData = True
        self.options.plotPredictedData = True
        self.options.showDataPoint = False
        self.options.showFittedPoint = False
        self.options.showConfidenceInterval = True
        self.options.dataColor = 'black'
        self.options.filteredColor = 'blue'
        self.options.predictedColor = 'green'
        self.options.smoothedColor = 'red'
        self.options.separatePlot = True
        self.options.confidence = 0.95
        self.options.intervalType = 'ribbon'


    # plot the result according to the options
    def plot(self, name='main'):
        """ The main plot function. The dlmPlot and the matplotlib will only be loaded
        when necessary.

        Args:
            name: component to plot. Default to 'main', in which we plot the
                  filtered time series. If a component name is given
                  It plots the mean of the component, i.e., the observed value
                  that attributes to that particular component, which equals to
                  evaluation * latent states for that particular component.

        """

        # load the library only when needed
        # import pydlm.plot.dlmPlot as dlmPlot
        self.loadPlotLibrary()

        if self.time is None:
            time = range(len(self.data))
        else:
            time = self.time

        # initialize the figure
        dlmPlot.plotInitialize()

        # change option setting if some results are not available
        if not self.initialized:
            raise NameError('The model must be constructed and' +
                            ' fitted before ploting.')

        # check the filter status and automatically turn off bad plots
        self._checkPlotOptions()

        # plot the main time series after filtering
        if name == 'main':
            # if we just need one plot
            if self.options.separatePlot is not True:
                dlmPlot.plotInOneFigure(time=time,
                                        data=self.data,
                                        result=self.result,
                                        options=self.options)
            # otherwise, we plot in multiple figures
            else:
                dlmPlot.plotInMultipleFigure(time=time,
                                             data=self.data,
                                             result=self.result,
                                             options=self.options)

        # plot the component after filtering
        elif self._checkComponent(name):
            # create the data for ploting
            data = {}
            if self.options.plotFilteredData:
                data['filteredMean'] = self.getMean(
                    filterType='forwardFilter', name=name)
                data['filteredVar'] = self.getVar(
                    filterType='forwardFilter', name=name)

            if self.options.plotSmoothedData:
                data['smoothedMean'] = self.getMean(
                    filterType='backwardSmoother', name=name)
                data['smoothedVar'] = self.getVar(
                    filterType='backwardSmoother', name=name)

            if self.options.plotPredictedData:
                data['predictedMean'] = self.getMean(
                    filterType='predict', name=name)
                data['predictedVar'] = self.getVar(
                    filterType='predict', name=name)

            if len(data) == 0:
                raise NameError('Nothing is going to be drawn, due to ' +
                                'user choices.')
            data['name'] = name
            dlmPlot.plotComponent(time=time,
                                  data=data,
                                  result=self.result,
                                  options=self.options)
        dlmPlot.plotout()


    def plotCoef(self, name, dimensions=None):
        """ Plot function for the latent states (coefficents of dynamic
        component).

        Args:
            name: the name of the component to plot.
                  It plots the latent states for the component. If dimension of
                  the given component is too high, we truncate
                  to the first five. Or the user can supply the ideal
                  dimensions for plot in the dimensions parameter.
            dimensions: dimensions will be used
                        as the indexes to plot within that component latent
                        states.
        """
        # load the library only when needed
        # import pydlm.plot.dlmPlot as dlmPlot
        self.loadPlotLibrary()

        # provide a fake time for plotting
        if self.time is None:
            time = range(len(self.data))
        else:
            time = self.time

        # change option setting if some results are not available
        if not self.initialized:
            raise NameError('The model must be constructed and' +
                            ' fitted before ploting.')

        # check the filter status and automatically turn off bad plots
        self._checkPlotOptions()

        # plot the latent states for a given component
        if self._checkComponent(name):

            # find its coordinates in the latent state
            indx = self.builder.componentIndex[name]
            # get the real latent states
            coordinates = range(indx[0], (indx[1] + 1))
            # if user supplies the dimensions
            if dimensions is not None:
                coordinates = [coordinates[i] for i in dimensions]

            # otherwise, if there are too many latent states, we
            # truncated to the first five.
            elif len(coordinates) > 5:
                coordinates = coordinates[:5]

            dlmPlot.plotLatentState(time=time,
                                    coordinates=coordinates,
                                    result=self.result,
                                    options=self.options,
                                    name=name)
        else:
            raise NameError('No such component.')

        dlmPlot.plotout()


    def plotPredictN(self, N=1, date=None, featureDict=None):
        """
        Function to plot the N-day prediction results.

        The input is the same as `dlm.predictN`. For details,
        please refer to that function.

        Args:
            N:    The length of days to predict.
            date: The index when the prediction based on. Default to the
                  last day.
            FeatureDict: The feature set for the dynamic Components, in a form
                  of {"component_name": feature}, where the feature must have
                  N elements of feature vectors. If the featureDict is not
                  supplied, then the algo reuses those stored in the dynamic
                  components. For dates beyond the last day, featureDict must
                  be supplied.
        """
        if date is None:
            date = self.n - 1

        # load the library only when needed
        # import pydlm.plot.dlmPlot as dlmPlot
        self.loadPlotLibrary()

        # provide a fake time for plotting
        if self.time is None:
            time = range(len(self.data))
        else:
            time = self.time

        # change option setting if some results are not available
        if not self.initialized:
            raise NameError('The model must be constructed and' +
                            ' fitted before ploting.')

        # check the filter status and automatically turn off bad plots
        self._checkPlotOptions()

        predictedTimeRange = range(date, date + N)
        predictedData, predictedVar = self.predictN(
            N=N, date=date, featureDict=featureDict)
        dlmPlot.plotPrediction(
            time=time, data=self.data,
            predictedTime=predictedTimeRange,
            predictedData=predictedData,
            predictedVar=predictedVar,
            options=self.options)

        dlmPlot.plotout()



    def loadPlotLibrary(self):
        if not self.plotLibLoaded:
            global dlmPlot
            import pydlm.plot.dlmPlot as dlmPlot
            self.plotLibLoaded = True



"""
=========================================================================

Code for the abstract component

=========================================================================

This piece of code provide the basic building block for the dynamic linear model.
It provide the fundamental struture for all model components. We implement different
components based on this abstract class.

"""
from abc import ABCMeta, abstractmethod

# We define an abstract class which can further be used
# to create different types of model components, inclusing
# trend, seasonality and other structures


class component:
    """ The abstract class provides the basic structure for all model components
    
    Methods:
        createEvaluation: create the initial evaluation matrix
        createTransition: create the initial transition matrix
        createCovPrior: create a simple prior covariance matrix
        createMeanPrior: create a simple prior latent state
        checkDimensions: if user supplies their own covPrior and meanPrior, this can 
                         be used to check if the dimension matches
    
    """
    __metaclass__ = ABCMeta

    def __init__(self):
        """All members that need to be initialized."""
        self.d = None
        self.name = None
        self.componentType = None
        self.discount = None
        self.evaluation = None
        self.transition = None
        self.covPrior = None
        self.meanPrior = None

    def __eq__(self, other):
        """Define equal method used for == comparison"""
        if not isinstance(other, component):
            return NotImplemented
        else:
            return (self.equalOrNone(self.d, other.d) and
                    self.equalOrNone(self.name, other.name) and
                    self.equalOrNone(self.componentType, other.componentType) and
                    self.npEqualOrNone(self.discount, other.discount) and
                    self.npEqualOrNone(self.evaluation, other.evaluation) and
                    self.npEqualOrNone(self.transition, other.transition) and
                    self.npEqualOrNone(self.covPrior, other.covPrior) and
                    self.npEqualOrNone(self.meanPrior, other.meanPrior))


    # define the evaluation matrix for the component
    @abstractmethod
    def createEvaluation(self): pass
    """ Create the evaluation matrix

    """


    # define the transition matrix for the component
    @abstractmethod
    def createTransition(self): pass
    """ Create the transition matrix

    """


    # define the prior distribution for the covariance for the component
    @abstractmethod
    def createCovPrior(self): pass
    """ Create the prior covariance matrix for the latent states

    """


    # define the prior distribution for the mean vector for the component
    @abstractmethod
    def createMeanPrior(self): pass
    """ Create the prior latent state
    
    """


    # check the matrix dimensions in case user supplied matrices are wrong
    @abstractmethod
    def checkDimensions(self): pass
    """ Check the dimensionality of the state and covariance

    """
    
    def equalOrNone(self, a, b):
        """Check if a and b are equal or both are None"""
        return (a is None and b is None) or a == b

    def npEqualOrNone(self, a, b):
        """Check if a and b are equal or both are None for NP arrays"""
        return (a is None and b is None) or (a == b).all()



"""
===============================================================================

The code for all hidden method for the class dlm

===============================================================================

This piece of code include all the hidden methods and members of the class dlm.
It provides the basic modeling, filtering, forecasting and smoothing of a dlm.

"""
from pydlm.base.kalmanFilter import kalmanFilter
from pydlm.modeler.builder import builder

from numpy import var
import logging

# this class defines the basic functionalities for dlm, which is not supposed
# to be used by the user. Most functionality in the main dlm will be
# constructed by using the hidden functions in this class


class _dlm(object):
    """ _dlm includes all basic functions for building dlm model.

    Attributes:
        data: the observed time series data
        n: the length of the time series data
        result: the inner class that records the filtered and smoothed results
        builder: the @builder that is used for providing the modeling
        functionality. For details please refer to @builder
        Filter: the filter used for filtering time series data using the model.
                For details please refer to @kalmanFilter
        initialized: indicates whether the dynamic linear model has been
                     initialized
        options: model options, including initial guess of the observational
                 variance.
                 More is going to be added (plot options and shrinkage options)
        time: the time label, used for plotting


    Methods:
        _initialize: initialize the dlm (builder and kalmanFilter)
        _forwardFilter: run forward filter for a specific start and end date
        _backwardSmoother: run backward smooth for a specific start and end
                           date
        _predictInSample: predict the latent state and observation for a given
                          period of time (deprecated)
        _resetModelStatus: reset the model status to its prior status
        _setModelStatus: set the model status to a specific date
        _defaultOptions: a class to store and set default options
        _result: a class to store the results
        _copy: copy the result from the model to the _result class
        _reverseCopy: copy the result from the _result class to the model
        _checkFeatureSize: check whether the features's n matches the data's n
        _checkComponent: check whether a component is in dlm
        _checkPlotOptions: set the correct options according to the fit
        _checkAndGetWorkingDates: get the correct filtering dates
    """


    # define the basic members
    # initialize the result
    def __init__(self, data, **options):

        self.data = list(data)
        # padded_data is used by auto regressor. It is the raw data with missing value
        # replaced by forward filter results. (Missing value include the out of scope
        # predictions.
        self.options = self._defaultOptions(**options)
        self._logger = self.options.logger
        self.builder = builder(self._logger)
        self.padded_data = self.data
        self.n = len(data)
        self.result = None
        self.Filter = None
        self.initialized = False
        self.time = None


    # an inner class to store all options
    class _defaultOptions(object):
        """ All plotting and fitting options

        """
        def __init__(self, **kwargs):
            self.noise = kwargs.get('noise', 1.0)
            # Stable mode is basically doing rolling window refitting. Use with caution.
            self.stable = kwargs.get('stable', False)
            self.innovationType = kwargs.get('component', 'component')

            self.plotOriginalData = kwargs.get('plotOriginalData', True)
            self.plotFilteredData = kwargs.get('plotFilteredData', True)
            self.plotSmoothedData = kwargs.get('plotSmoothedData', True)
            self.plotPredictedData = kwargs.get('plotPredictedData', True)
            self.showDataPoint = kwargs.get('showDataPoint', False)
            self.showFittedPoint = kwargs.get('showFittedPoint', False)
            self.showConfidenceInterval = kwargs.get('showConfidenceInterval', True)
            self.dataColor = kwargs.get('dataColor', 'black')
            self.filteredColor = kwargs.get('filteredColor', 'blue')
            self.predictedColor = kwargs.get('predictedColor', 'green')
            self.smoothedColor = kwargs.get('smoothedColor', 'red')
            self.separatePlot = kwargs.get('seperatePlot', True)
            self.confidence = kwargs.get('confidence', 0.95)
            self.intervalType = kwargs.get('intervalType', 'ribbon')
            self.useAutoNoise = kwargs.get('useAutoNoise', False)
            self.logger = kwargs.get('logger', self._getDefaultLogger())

        def _getDefaultLogger(self):
            """ Get default logger """
        
            logging.basicConfig()
            logger = logging.getLogger('pydlm')
            logger.setLevel(logging.INFO)
            return logger




    # an inner class to store all results
    class _result(object):
        """ Class to store the results

        """
        # class level (static) variables to record all names
        records = ['filteredObs', 'predictedObs', 'smoothedObs',
                   'filteredObsVar',
                   'predictedObsVar', 'smoothedObsVar', 'noiseVar',
                   'df',
                   'filteredState', 'predictedState', 'smoothedState',
                   'filteredCov', 'predictedCov', 'smoothedCov']


        # quantites to record the result
        def __init__(self, n):

            # initialize all records to be [None] * n
            for variable in self.records:
                setattr(self, variable, [None] * n)

            # record the dates that have been filtered
            self.filteredSteps = [0, -1]
            # record the dates that have been smoothed
            self.smoothedSteps = [0, -1]
            # record the last used filterType
            self.filteredType = None
            # record the current prediction status in the form of
            # [start date, current date, [predictedObs1, predictedObs2,...]]
            self.predictStatus = None
            # One day ahead prediction squared error


        # extend the current record by n blocks
        def _appendResult(self, n):
            for variable in self.records:
                getattr(self, variable).extend([None] * n)


        # pop out a specific date
        def _popout(self, date):
            for variable in self.records:
                getattr(self, variable).pop(date)


    # initialize the builder
    def _initialize(self):
        """ Initialize the model: initialize builder and filter.

        """
        self._autoNoise()
        self.builder.initialize(noise=self.options.noise, data=self.padded_data)
        self.Filter = kalmanFilter(discount=self.builder.discount,
                                   updateInnovation=self.options.innovationType,
                                   index=self.builder.componentIndex)
        self.result = self._result(self.n)
        self.initialized = True


    def _initializeFromBuilder(self, exported_builder):
        self.builder.initializeFromBuilder(data=self.data, exported_builder=exported_builder)
        self.Filter = kalmanFilter(discount=self.builder.discount,
                                   updateInnovation=self.options.innovationType,
                                   index=self.builder.componentIndex)
        self.result = self._result(self.n)
        self.initialized = True


    def _autoNoise(self):
        """ Auto initialize the noise parameter if options.useAutoNoise
        is true.

        """
        if self.options.useAutoNoise:
            trimmed_data = [x for x in self.data if x is not None]
            self.options.noise = min(var(trimmed_data), 1)


   # use the forward filter to filter the data
    # start: the place where the filter started
    # end: the place where the filter ended
    # save: the index for dates where the filtered results should be saved,
    #       could be 'all' or 'end'
    # isForget: indicate where the filter should use the previous state as
    #         prior or just use the prior information from builder
    def _forwardFilter(self,
                       start=0,
                       end=None,
                       save='all',
                       ForgetPrevious=False,
                       renew=False):
        """ Running forwardFilter for the data for a given start and end date

        Args:
            start: the start date
            end: the end date (default to the last day of the chain)
            save: indicate the dates of which the result needs to be saved for.
                  'all' stands for (start, end), otherwise an integer between
                  start and end.
            ForgetPrevious: indicate whether the fitering should start from
                            the prior status or the previous date that has
                            been filtered.
                            (used for rolling window filtering, see @dlm)
            renew: if true, filter will refit certain days when the
                   chain gets
                   too long to add numerical stability, the length of the chain
                   is determined by the information carried on. For example,
                   when discount = 0.9, any days that are 65 days ago together
                   only carry information 1%, so we ignore
                   these days and refit the model to aid stability.
        """
        # the default value for end
        if end is None:
            end = self.n - 1

        # to see if the ff need to run or not
        if start > end:
            return None

        # first we need to initialize the model to the correct status
        # if the start point is 0 or we want to forget the previous result
        if start == 0 or ForgetPrevious:
            self._resetModelStatus()

        # otherwise we use the result on the previous day as the prior
        else:
            if start > self.result.filteredSteps[1] + 1:
                raise ValueError('The data before start date has'
                                ' yet to be filtered! Otherwise set'
                                ' ForgetPrevious to be True. Check the'
                                ' <filteredSteps> in <result> object.')
            self._setModelStatus(date=start - 1)

        # we run the forward filter sequentially
        lastRenewPoint = start  # record the last renew point
        for step in range(start, end + 1):

            # first check whether we need to update evaluation or not
            if len(self.builder.dynamicComponents) > 0 or \
               len(self.builder.automaticComponents) > 0:
                self.builder.updateEvaluation(step, self.padded_data)

            # check if rewnew is needed
            if renew and step - lastRenewPoint > self.builder.renewTerm \
               and self.builder.renewTerm > 0.0:
                # we renew the state of the day
                self._resetModelStatus()
                for innerStep in range(step - int(self.builder.renewTerm),
                                       step):
                    self.Filter.forwardFilter(self.builder.model,
                                              self.data[innerStep])
                lastRenewPoint = step

            # then we use the updated model to filter the state
            self.Filter.forwardFilter(self.builder.model, self.data[step])

            # extract the result and record
            if save == 'all' or save == step:
                self._copy(model=self.builder.model,
                           result=self.result,
                           step=step,
                           filterType='forwardFilter')


    # use the backward smooth to smooth the state
    # start: the last date of the backward filtering chain
    # days: number of days to go back from start
    def _backwardSmoother(self, start=None, days=None, ignoreFuture=False):
        """ Backward smooth over filtered results for a specific start
            and number of days

        Args:
            start: the start date
            days: number of days to be smoothed starting from start towards
                  zero
            ignoreFuture: indicate whether the smoothed should start as if the
                          future data was not observed or using the future data
                          as the initial smoothing status.
        """
        # the default start date is the most recent date
        if start is None:
            start = self.n - 1

        # the default backward days number is the total length
        if days is None:
            end = 0
        else:
            end = max(start - days + 1, 0)

        # the forwardFilter has to be run before the smoother
        if self.result.filteredSteps[1] < start:
            raise valueError('The last day has to be filtered before smoothing!'
                             'check the <filteredSteps> in <result> object.')

        # and we record the most recent day which does not need to be smooth
        if start == self.n - 1 or ignoreFuture is True:
            self.result.smoothedState[start] = self.result.filteredState[start]
            self.result.smoothedObs[start] = self.result.filteredObs[start]
            self.result.smoothedCov[start] = self.result.filteredCov[start]
            self.result.smoothedObsVar[start] \
                = self.result.filteredObsVar[start]
            self.builder.model.noiseVar = self.result.noiseVar[start]
            start -= 1
        else:
            self.builder.model.noiseVar = self.result.noiseVar[self.n - 1]

        # empty smoothing chain, return None
        if start < end:
            return None

        # insert the previous smoothed dates
        self.builder.model.state = self.result.smoothedState[start + 1]
        self.builder.model.sysVar = self.result.smoothedCov[start + 1]

        # we smooth the result sequantially from start - 1 to end
        dates = list(range(end, start + 1))
        dates.reverse()
        for day in dates:
            # we first update the model to be correct status before smooth
            self.builder.model.prediction.state \
                = self.result.predictedState[day + 1]
            self.builder.model.prediction.sysVar \
                = self.result.predictedCov[day + 1]

            if len(self.builder.dynamicComponents) > 0 or \
               len(self.builder.automaticComponents) > 0:
                self.builder.updateEvaluation(day, self.padded_data)

            # then we use the backward filter to filter the result
            self.Filter.backwardSmoother(
                model=self.builder.model,
                rawState=self.result.filteredState[day],
                rawSysVar=self.result.filteredCov[day])

            # extract the result
            self._copy(model=self.builder.model,
                       result=self.result,
                       step=day,
                       filterType='backwardSmoother')


    # Forecast the result based on filtered chains
    def _predictInSample(self, date, days=1):
        """ Predict the model's status based on the model of a specific day

        Args:
            date: the date the prediction is based on
            day: number of days forward that need to be predicted.

        Returns:
            A tuple. (Predicted observation, variance of the predicted
            observation)

        """

        if date + days > self.n - 1:
            raise ValueError('The range is out of sample.')

        predictedObs = [None] * days
        predictedObsVar = [None] * days
        # reset the date to the date we are interested in
        self._setModelStatus(date=date)
        self.builder.model.prediction.step = 0
        for i in range(1, days):
            # update the evaluation vector
            if len(self.builder.dynamicComponents) > 0 or \
               len(self.builder.automaticComponents) > 0:
                self.builder.updateEvaluation(date + i, self.padded_data)

            self.Filter.predict(self.builder.model)
            predictedObs[i - 1] = self.builder.model.prediction.obs
            predictedObsVar[i - 1] = self.builder.model.prediction.obsVar

        return (predictedObs, predictedObsVar)

# =========================== model helper function ==========================


    # to set model to a specific date
    def _setModelStatus(self, date=0):
        """ Set the model status to a specific date (the date mush have been filtered)

        """
        if date < self.result.filteredSteps[0] or \
           date > self.result.filteredSteps[1]:
            raise ValueError('The date has yet to be filtered yet. ' 
                             'Check the <filteredSteps> in <result> object.')

        self._reverseCopy(model=self.builder.model,
                          result=self.result,
                          step=date)
        if len(self.builder.dynamicComponents) > 0 or \
           len(self.builder.automaticComponents) > 0:
            self.builder.updateEvaluation(date, self.padded_data)


    # reset model to initial status
    def _resetModelStatus(self):
        """ Reset the model to the prior status

        """
        self.builder.model.state = self.builder.statePrior
        self.builder.model.sysVar = self.builder.sysVarPrior
        self.builder.model.noiseVar = self.builder.noiseVar
        self.builder.model.df = self.builder.initialDegreeFreedom
        self.builder.model.initializeObservation()


    # a function used to copy result from the model to the result
    def _copy(self, model, result, step, filterType):
        """ Copy result from the model to _result class

        """

        if filterType == 'forwardFilter':
            result.filteredObs[step] = model.obs
            result.predictedObs[step] = model.prediction.obs
            result.filteredObsVar[step] = model.obsVar
            result.predictedObsVar[step] = model.prediction.obsVar
            result.filteredState[step] = model.state
            result.predictedState[step] = model.prediction.state
            result.filteredCov[step] = model.sysVar
            result.predictedCov[step] = model.prediction.sysVar
            result.noiseVar[step] = model.noiseVar
            result.df[step] = model.df
            # pad missing value with filtered result
            if self.data[step] is None:
                self.padded_data[step] = result.filteredObs[step]

        elif filterType == 'backwardSmoother':
            result.smoothedState[step] = model.state
            result.smoothedObs[step] = model.obs
            result.smoothedCov[step] = model.sysVar
            result.smoothedObsVar[step] = model.obsVar


    def _reverseCopy(self, model, result, step):
        """ Copy result from _result class to the model

        """

        model.obs = result.filteredObs[step]
        model.prediction.obs = result.predictedObs[step]
        model.obsVar = result.filteredObsVar[step]
        model.prediction.obsVar = result.predictedObsVar[step]
        model.state = result.filteredState[step]
        model.prediction.state = result.predictedState[step]
        model.sysVar = result.filteredCov[step]
        model.prediction.sysVar = result.predictedCov[step]
        model.noiseVar = result.noiseVar[step]
        model.df = result.df[step]


    # check if the data size matches the dynamic features
    def _checkFeatureSize(self):
        """ Check features's n matches the data's n

        """
        if len(self.builder.dynamicComponents) > 0:
            for name in self.builder.dynamicComponents:
                if self.builder.dynamicComponents[name].n != self.n:
                    raise ValueError(f"The data size of dlm and { name } does not match")


    def _1DmatrixToArray(self, arrayOf1dMatrix):
        """ Change an array of 1 x 1 matrix to normal array.

        """
        return [item.tolist()[0][0] for item in arrayOf1dMatrix]


    # function to judge whether a component is in the model
    def _checkComponent(self, name):
        """ Check whether a component is contained by the dlm.

        Args:
            name: the name of the component

        Returns:
            True or error.
        """
        if name in self.builder.staticComponents or \
           name in self.builder.dynamicComponents or \
           name in self.builder.automaticComponents:
            return True
        else:
            raise ValueError('No such component.')


    # function to fetch a component
    def _fetchComponent(self, name):
        """ Get the component if the componeng is in the dlm

        Args:
            name: the name of the component

        Returns:
            The component or error.
        """
        if name in self.builder.staticComponents:
            comp = self.builder.staticComponents[name]
        elif name in self.builder.dynamicComponents:
            comp = self.builder.dynamicComponents[name]
        elif name in self.builder.automaticComponents:
            comp = self.builder.automaticComponents[name]
        else:
            raise ValueError('No such component.')
        return comp


    # check start and end dates that has been filtered on
    def _checkAndGetWorkingDates(self, filterType):
        """ Check the filter status and return the dates that have
        been filtered according to the filterType.

        Args:
            filterType: the type of the filter to check. Could be
                       "forwardFilter", "backwardSmoother" and "predict"

        Returns:
            a tuple of (start_date, end_date)
        """
        if filterType == 'forwardFilter' or filterType == 'predict':
            if self.result.filteredSteps != [0, self.n - 1]:
                self._logger.info('The fitlered dates are from ' +
                      str(self.result.filteredSteps[0]) +
                      ' to ' + str(self.result.filteredSteps[1]))
            start = self.result.filteredSteps[0]
            end = self.result.filteredSteps[1]

        elif filterType == 'backwardSmoother':
            if self.result.smoothedSteps != [0, self.n - 1]:
                self._logger.info('The smoothed dates are from ' +
                      str(self.result.smoothedSteps[0]) +
                      ' to ' + str(self.result.smoothedSteps[1]))
            start = self.result.smoothedSteps[0]
            end = self.result.smoothedSteps[1]

        else:
            raise ValueError('Incorrect filter type.')

        return (start, end)


    # check the filter status and automatic turn off some plots
    def _checkPlotOptions(self):
        """ Check the filter status and determine the plot options.

        """
        # adjust the option according to the filtering status
        if self.result.filteredSteps[1] == -1:
            self.options.plotFilteredData = False
            self.options.plotPredictedData = False

        if self.result.smoothedSteps[1] == -1:
            self.options.plotSmoothedData = False


    def setLoggingLevel(self, level='INFO'):
        """ The level of logs that should be printed.

        """
        if level == 'DEBUG':
            self._logger.setLevel(logging.DEBUG)
        elif level == 'INFO':
            self._logger.setLevel(logging.INFO)
        elif level == 'WARNING':
            self._logger.setLevel(logging.WARNING)
        elif level == 'CRITICAL':
            self._logger.setLevel(logging.CRITICAL)
        else:
            raise ValueError(f"Log at {level} is not supported.")

    def getLoggingLevel(self):
        """ Get the level of logger."""

        return logging.getLevelName(self._logger.getEffectiveLevel())



from copy import deepcopy
from pydlm.base.tools import getInterval
from pydlm.access._dlmGet import _dlmGet


class dlmAccessModule(_dlmGet):
    """ A dlm module for all the access methods
    """


    def getAll(self):
        """ get all the _result class which contains all results

        Returns:
            The @result object containing all computed results.

        """
        return deepcopy(self.result)


    def getMean(self, filterType='forwardFilter', name='main'):
        """ get mean for data or component.

        If the working dates are not (0, self.n - 1),
        then a warning will prompt stating the actual filtered dates.

        Args:
            filterType: the type of mean to be returned. Could be
                        'forwardFilter', 'backwardSmoother', and 'predict'.
                        Default to 'forwardFilter'.
            name: the component to get mean. When name = 'main', then it
                  returns the filtered mean for the time series. When
                  name = some component's name, then it returns the filtered
                  mean for that component. Default to 'main'.

        Returns:
            A list of the time series observations based on the choice

        """
        # get the working date
        start, end = self._checkAndGetWorkingDates(filterType=filterType)
        end += 1 # To get the result for the last date.
        # get the mean for the fitlered data
        if name == 'main':
            # get out of the matrix form
            if filterType == 'forwardFilter':
                return self._1DmatrixToArray(
                    self.result.filteredObs[start:end])
            elif filterType == 'backwardSmoother':
                return self._1DmatrixToArray(
                    self.result.smoothedObs[start:end])
            elif filterType == 'predict':
                return self._1DmatrixToArray(
                    self.result.predictedObs[start:end])
            else:
                raise NameError('Incorrect filter type.')

        # get the mean for the component
        self._checkComponent(name)
        return self._getComponentMean(name=name,
                                      filterType=filterType,
                                      start=start, end=(end - 1))


    def getVar(self, filterType='forwardFilter', name='main'):
        """ get the variance for data or component.

        If the filtered dates are not (0, self.n - 1),
        then a warning will prompt stating
        the actual filtered dates.

        Args:
            filterType: the type of variance to be returned. Could be
                        'forwardFilter', 'backwardSmoother', and 'predict'.
                        Default to 'forwardFilter'.
            name: the component to get variance. When name = 'main', then it
                  returns the filtered variance for the time series. When
                  name = some component's name, then it returns the filtered
                  variance for that component. Default to 'main'.

        Returns:
            A list of the filtered variances based on the choice.

        """
        # get the working date
        start, end = self._checkAndGetWorkingDates(filterType=filterType)
        end += 1
        # get the variance for the time series data
        if name == 'main':
            # get out of the matrix form
            if filterType == 'forwardFilter':
                return self._1DmatrixToArray(
                    self.result.filteredObsVar[start:end])
            elif filterType == 'backwardSmoother':
                return self._1DmatrixToArray(
                    self.result.smoothedObsVar[start:end])
            elif filterType == 'predict':
                return self._1DmatrixToArray(
                    self.result.predictedObsVar[start:end])
            else:
                raise NameError('Incorrect filter type.')

        # get the variance for the component
        self._checkComponent(name)
        return self._getComponentVar(name=name, filterType=filterType,
                                     start=start, end=(end - 1))


    def getResidual(self, filterType='forwardFilter'):
        """ get the residuals for data after filtering or smoothing.

        If the working dates are not (0, self.n - 1),
        then a warning will prompt stating the actual filtered dates.

        Args:
            filterType: the type of residuals to be returned. Could be
                        'forwardFilter', 'backwardSmoother', and 'predict'.
                        Default to 'forwardFilter'.

        Returns:
            A list of residuals based on the choice

        """
        # get the working date
        start, end = self._checkAndGetWorkingDates(filterType=filterType)
        end += 1 # To get the result for the last date.
        # get the mean for the fitlered data
        # get out of the matrix form
        if filterType == 'forwardFilter':
            return self._1DmatrixToArray(
                [self.data[i] - self.result.filteredObs[i]
                 for i in range(start, end)])
        elif filterType == 'backwardSmoother':
            return self._1DmatrixToArray(
                [self.data[i] - self.result.smoothedObs[i]
                 for i in range(start, end)])
        elif filterType == 'predict':
            return self._1DmatrixToArray(
                [self.data[i] - self.result.predictedObs[i]
                 for i in range(start, end)])
        else:
            raise NameError('Incorrect filter type.')


    def getInterval(self, p=0.95, filterType='forwardFilter', name='main'):
        """ get the confidence interval for data or component.

        If the filtered dates are not
        (0, self.n - 1), then a warning will prompt stating the actual
        filtered dates.

        Args:
            p: The confidence level.
            filterType: the type of CI to be returned. Could be
                        'forwardFilter', 'backwardSmoother', and 'predict'.
                        Default to 'forwardFilter'.
            name: the component to get CI. When name = 'main', then it
                  returns the confidence interval for the time series. When
                  name = some component's name, then it returns the confidence
                  interval for that component. Default to 'main'.

        Returns:
            A tuple with the first element being a list of upper bounds
            and the second being a list of the lower bounds.

        """
        # get the working date
        start, end = self._checkAndGetWorkingDates(filterType=filterType)
        end += 1
        # get the mean and the variance for the time series data
        if name == 'main':
            # get out of the matrix form
            if filterType == 'forwardFilter':
                compMean = self._1DmatrixToArray(
                    self.result.filteredObs[start:end])
                compVar = self._1DmatrixToArray(
                    self.result.filteredObsVar[start:end])
            elif filterType == 'backwardSmoother':
                compMean = self._1DmatrixToArray(
                    self.result.smoothedObs[start:end])
                compVar = self._1DmatrixToArray(
                    self.result.smoothedObsVar[start:end])
            elif filterType == 'predict':
                compMean = self._1DmatrixToArray(
                    self.result.predictedObs[start:end])
                compVar = self._1DmatrixToArray(
                    self.result.predictedObsVar[start:end])
            else:
                raise NameError('Incorrect filter type.')

        # get the mean and variance for the component
        else:
            self._checkComponent(name)
            compMean = self._getComponentMean(name=name,
                                              filterType=filterType,
                                              start=start, end=(end - 1))
            compVar = self._getComponentVar(name=name,
                                            filterType=filterType,
                                            start=start, end=(end - 1))

        # get the upper and lower bound
        upper, lower = getInterval(compMean, compVar, p)
        return (upper, lower)


    def getLatentState(self, filterType='forwardFilter', name='all'):
        """ get the latent states for different components and filters.

        If the filtered dates are not (0, self.n - 1),
        then a warning will prompt stating the actual filtered dates.

        Args:
            filterType: the type of latent states to be returned. Could be
                        'forwardFilter', 'backwardSmoother', and 'predict'.
                        Default to 'forwardFilter'.
            name: the component to get latent state. When name = 'all', then it
                  returns the latent states for the time series. When
                  name = some component's name, then it returns the latent
                  states for that component. Default to 'all'.

        Returns:
            A list of lists, standing for the latent states given
            the different choices.

        """
        # get the working dates
        start, end = self._checkAndGetWorkingDates(filterType=filterType)
        end += 1
        # to return the full latent states
        if name == 'all':
            if filterType == 'forwardFilter':
                return list(map(lambda x: x if x is None
                                else self._1DmatrixToArray(x),
                                self.result.filteredState[start:end]))
            elif filterType == 'backwardSmoother':
                return list(map(lambda x: x if x is None
                                else self._1DmatrixToArray(x),
                                self.result.smoothedState[start:end]))
            elif filterType == 'predict':
                return list(map(lambda x: x if x is None
                                else self._1DmatrixToArray(x),
                                self.result.smoothedState[start:end]))
            else:
                raise NameError('Incorrect filter type.')

        # to return the latent state for a given component
        self._checkComponent(name)
        return list(map(lambda x: x if x is None else self._1DmatrixToArray(x),
                        self._getLatentState(name=name, filterType=filterType,
                                             start=start, end=(end - 1))))


    def getLatentCov(self, filterType='forwardFilter', name='all'):
        """ get the error covariance for different components and
        filters.

        If the filtered dates are not (0, self.n - 1),
        then a warning will prompt stating the actual filtered dates.

        Args:
            filterType: the type of latent covariance to be returned. Could be
                        'forwardFilter', 'backwardSmoother', and 'predict'.
                        Default to 'forwardFilter'.
            name: the component to get latent cov. When name = 'all', then it
                  returns the latent covariance for the time series. When
                  name = some component's name, then it returns the latent
                  covariance for that component. Default to 'all'.

        Returns:
            A list of numpy matrices, standing for the filtered latent
            covariance.

        """
        # get the working dates
        start, end = self._checkAndGetWorkingDates(filterType=filterType)
        end += 1
        # to return the full latent covariance
        if name == 'all':
            if filterType == 'forwardFilter':
                return self.result.filteredCov[start:end]
            elif filterType == 'backwardSmoother':
                return self.result.smoothedCov[start:end]
            elif filterType == 'predict':
                return self.result.smoothedCov[start:end]
            else:
                raise NameError('Incorrect filter type.')

        # to return the latent covariance for a given component
        self._checkComponent(name)
        return self._getLatentCov(name=name, filterType=filterType,
                                  start=start, end=(end - 1))
